import json

from django.core.exceptions import ValidationError
from django.db.models import Count, ExpressionWrapper, FloatField, Sum
from django.db.models.functions import Coalesce, NullIf
from django.forms import ChoiceField
from django.shortcuts import get_object_or_404
from django.urls import reverse
from django_datatables.datatables import DatatableTable
from django_menus.menu import MenuItem
from django_modals.fields import FieldEx
from django_modals.form_helpers import HorizontalNoEnterHelper
from django_modals.helper import show_modal
from django_modals.modals import Modal, ModelFormModal
from django_modals.processes import PERMISSION_OFF, PROCESS_EDIT_DELETE
from django_modals.widgets.select2 import Select2Multiple
from django_modals.widgets.widgets import Toggle

from advanced_report_builder.column_types import NUMBER_FIELDS
from advanced_report_builder.columns import ReportBuilderNumberColumn
from advanced_report_builder.exceptions import ReportError
from advanced_report_builder.globals import (
    ANNOTATION_CHOICE_AVERAGE_SUM_FROM_COUNT,
    ANNOTATION_CHOICE_SUM,
)
from advanced_report_builder.models import ReportQuery, ReportType, SingleValueReport
from advanced_report_builder.utils import get_report_builder_class
from advanced_report_builder.variable_date import VariableDate
from advanced_report_builder.views.charts_base import ChartBaseView
from advanced_report_builder.views.datatables.modal import (
    TableFieldForm,
    TableFieldModal,
)
from advanced_report_builder.views.datatables.utils import TableUtilsMixin
from advanced_report_builder.views.helpers import QueryBuilderModelForm
from advanced_report_builder.views.modals_base import (
    QueryBuilderModalBase,
    QueryBuilderModalBaseMixin,
)
from advanced_report_builder.views.query_modal.mixin import MultiQueryModalMixin


class SingleValueView(ChartBaseView):
    number_field = ReportBuilderNumberColumn
    template_name = 'advanced_report_builder/single_values/report.html'
    use_annotations = False

    def dispatch(self, request, *args, **kwargs):
        self.report = kwargs.get('report')
        self.chart_report = self.report.singlevaluereport
        return super().dispatch(request, *args, **kwargs)

    def _process_aggregations(self, fields, aggregations_type=ANNOTATION_CHOICE_SUM, divider=None):
        field = self.chart_report.field
        base_model = self.chart_report.get_base_model()
        report_builder_class = get_report_builder_class(model=base_model, report_type=self.chart_report.report_type)

        django_field, col_type_override, _, _ = self.get_field_details(
            base_model=base_model,
            field=field,
            report_builder_class=report_builder_class,
        )

        if isinstance(django_field, NUMBER_FIELDS) or col_type_override is not None and col_type_override.annotations:
            self.get_number_field(
                annotations_type=aggregations_type,
                append_annotation_query=False,
                index=0,
                data_attr={},
                table_field={'field': field, 'title': field},
                fields=fields,
                col_type_override=col_type_override,
                decimal_places=self.chart_report.decimal_places,
                convert_currency_fields=True,
                divider=divider,
            )
        else:
            raise ReportError('not a number field')

    def _get_count(self, fields):
        number_function_kwargs = {
            'aggregations': {'count': Count(1)},
            'field': 'count',
            'column_name': 'count',
            'options': {'calculated': True},
            'model_path': '',
        }

        field = self.number_field(**number_function_kwargs)
        fields.append(field)

    def get_percentage_field(
        self,
        fields,
        numerator_field_name,
        numerator_col_type_override,
        denominator_field_name,
        denominator_col_type_override,
        numerator_filter,
        decimal_places=0,
    ):
        if numerator_col_type_override:
            actual_numerator_field_name = numerator_col_type_override.field
        else:
            actual_numerator_field_name = numerator_field_name

        if denominator_col_type_override:
            actual_denominator_field_name = denominator_col_type_override.field
        else:
            actual_denominator_field_name = denominator_field_name

        number_function_kwargs = {}
        if decimal_places:
            number_function_kwargs = {'decimal_places': int(decimal_places)}

        new_field_name = f'{actual_numerator_field_name}_{actual_denominator_field_name}'

        if numerator_col_type_override is not None and numerator_col_type_override.annotations:
            if not isinstance(numerator_col_type_override.annotations, dict):
                raise ReportError('Unknown annotation type')

            annotations = list(numerator_col_type_override.annotations.values())[0]
            if numerator_filter and isinstance(annotations, Sum | Count):
                annotations.filter = numerator_filter
                numerator = Coalesce(annotations + 0.0, 0.0)
            else:
                numerator = Coalesce(annotations + 0.0, 0.0)
        else:
            if numerator_filter:
                numerator = Coalesce(
                    (Sum(actual_numerator_field_name, filter=numerator_filter) + 0.0),
                    0.0,
                )
            else:
                numerator = Coalesce((Sum(actual_numerator_field_name) + 0.0), 0.0)

        if denominator_col_type_override is not None and denominator_col_type_override.annotations:
            if not isinstance(denominator_col_type_override.annotations, dict):
                raise ReportError('Unknown annotation type')

            annotations = list(denominator_col_type_override.annotations.values())[0]

            denominator = Coalesce(annotations + 0.0, 0.0)
        else:
            denominator = Coalesce(Sum(actual_denominator_field_name) + 0.0, 0.0)

        number_function_kwargs['aggregations'] = {
            new_field_name: ExpressionWrapper((numerator / NullIf(denominator, 0)) * 100.0, output_field=FloatField())
        }
        field_name = new_field_name

        number_function_kwargs.update(
            {
                'field': field_name,
                'column_name': field_name,
                'options': {'calculated': True},
                'model_path': '',
            }
        )

        field = self.number_field(**number_function_kwargs)

        fields.append(field)

        return field_name

    def _process_percentage(self, fields):
        denominator_field = self.chart_report.field
        numerator_field = self.chart_report.numerator
        base_model = self.chart_report.get_base_model()
        report_builder_class = get_report_builder_class(model=base_model, report_type=self.chart_report.report_type)

        deno_django_field, denominator_col_type_override, _, _ = self.get_field_details(
            base_model=base_model,
            field=denominator_field,
            report_builder_class=report_builder_class,
        )
        if not isinstance(deno_django_field, NUMBER_FIELDS) and (
            denominator_col_type_override is not None and not denominator_col_type_override.annotations
        ):
            raise ReportError('denominator is not a number field')

        num_django_field, numerator_col_type_override, _, _ = self.get_field_details(
            base_model=base_model,
            field=denominator_field,
            report_builder_class=report_builder_class,
        )

        if not isinstance(num_django_field, NUMBER_FIELDS) and (
            numerator_col_type_override is not None and not numerator_col_type_override.annotations
        ):
            raise ReportError('numerator is not a number field')

        numerator_filter = None
        report_query = self.get_report_query(report=self.chart_report)
        if report_query:
            numerator_filter = self.process_filters(search_filter_data=report_query.extra_query)

        self.get_percentage_field(
            fields=fields,
            numerator_field_name=numerator_field,
            numerator_col_type_override=numerator_col_type_override,
            denominator_field_name=denominator_field,
            denominator_col_type_override=denominator_col_type_override,
            numerator_filter=numerator_filter,
            decimal_places=self.chart_report.decimal_places,
        )

    def _process_percentage_from_count(self, fields):
        report_query = self.get_report_query(report=self.chart_report)

        numerator_filter = None
        if report_query:
            numerator_filter = self.process_filters(search_filter_data=report_query.extra_query)

        if numerator_filter:
            numerator = Coalesce(Count(1, filter=numerator_filter) + 0.0, 0.0)
        else:
            numerator = Coalesce(Count(1) + 0.0, 0.0)
        denominator = Coalesce(Count(1) + 0.0, 0.0)
        a = (numerator / denominator) * 100.0

        number_function_kwargs = {
            'aggregations': {'count': a},
            'field': 'count',
            'column_name': 'count',
            'options': {'calculated': True},
            'model_path': '',
        }
        decimal_places = self.chart_report.decimal_places
        if decimal_places:
            number_function_kwargs['decimal_places'] = int(decimal_places)

        field = self.number_field(**number_function_kwargs)
        fields.append(field)

    def process_query_results(self, base_model, table):
        single_value_type = self.chart_report.single_value_type
        fields = []
        if single_value_type == SingleValueReport.SINGLE_VALUE_TYPE_COUNT:
            self._get_count(fields=fields)
        elif single_value_type == SingleValueReport.SINGLE_VALUE_TYPE_SUM:
            self._process_aggregations(fields=fields, aggregations_type=ANNOTATION_CHOICE_SUM)
        elif single_value_type == SingleValueReport.SINGLE_VALUE_TYPE_COUNT_AND_SUM:
            self._get_count(fields=fields)
            self._process_aggregations(fields=fields, aggregations_type=ANNOTATION_CHOICE_SUM)
        elif single_value_type == SingleValueReport.SINGLE_VALUE_TYPE_AVERAGE_SUM_FROM_COUNT:
            self._process_aggregations(
                fields=fields,
                aggregations_type=ANNOTATION_CHOICE_AVERAGE_SUM_FROM_COUNT,
            )
        elif single_value_type == SingleValueReport.SINGLE_VALUE_TYPE_AVERAGE_SUM_OVER_TIME:
            divider = self.get_period_divider(
                annotation_value_choice=self.chart_report.average_scale,
                start_date_type=self.chart_report.average_start_period,
                end_date_type=self.chart_report.average_end_period,
            )
            self._process_aggregations(fields=fields, aggregations_type=ANNOTATION_CHOICE_SUM, divider=divider)
        elif single_value_type == SingleValueReport.SINGLE_VALUE_TYPE_PERCENT:
            self._process_percentage(fields=fields)
        elif single_value_type == SingleValueReport.SINGLE_VALUE_TYPE_PERCENT_FROM_COUNT:
            self._process_percentage_from_count(fields=fields)
        return fields

    def set_prefix(self):
        self.table.prefix = self.chart_report.prefix
        if self.table.prefix:
            self.table.prefix += ' '

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        self.set_prefix()
        self.table.single_value = self.chart_report
        self.table.enable_links = self.kwargs.get('enable_links')
        self.table.datatable_template = 'advanced_report_builder/single_values/middle.html'
        self.table.breakdown_url = self.get_breakdown_url()
        context['single_value_report'] = self.chart_report
        return context

    def get_breakdown_slug(self):
        query_id = self.slug.get(f'query{self.report.pk}')
        if query_id is None and self.dashboard_report is not None:
            query_id = self.dashboard_report.report_query_id
        slug = f'pk-{self.table.single_value.id}-enable_links-{self.table.enable_links}'
        if query_id is not None:
            slug += f'-query{self.report.pk}-{query_id}'
        return slug

    def get_breakdown_url(self):
        if self.table.single_value.show_breakdown:
            slug = self.get_breakdown_slug()
            return show_modal(
                'advanced_report_builder:single_value_show_breakdown_modal',
                slug,
                href=True,
            )
        return None

    def pod_dashboard_view_menu(self):
        return []

    def edit_report_menu(self, request, chart_report_id, slug_str):
        return [
            MenuItem(
                f'advanced_report_builder:single_value_modal,pk-{chart_report_id}{slug_str}',
                menu_display='Edit',
                font_awesome='fas fa-pencil-alt',
                css_classes=['btn-primary'],
            ),
            *self.duplicate_menu(request=self.request, report_id=chart_report_id),
        ]


class SingleValueModal(MultiQueryModalMixin, QueryBuilderModalBase):
    template_name = 'advanced_report_builder/single_values/modal.html'
    process = PROCESS_EDIT_DELETE
    permission_delete = PERMISSION_OFF
    model = SingleValueReport
    show_order_by = False
    widgets = {
        'report_tags': Select2Multiple,
        'show_breakdown': Toggle(attrs={'data-onstyle': 'success', 'data-on': 'YES', 'data-off': 'NO'}),
    }

    form_fields = [
        'name',
        'notes',
        'report_type',
        'report_tags',
        ('single_value_type', {'label': 'Value type'}),
        ('numerator', {'label': 'Numerator field'}),
        'average_scale',
        'average_start_period',
        'average_end_period',
        'field',
        'prefix',
        'tile_colour',
        ('decimal_places', {'field_class': 'col-md-5 col-lg-3 input-group-sm'}),
        'show_breakdown',
        'breakdown_fields',
    ]

    def form_setup(self, form, *_args, **_kwargs):
        form.add_trigger(
            'single_value_type',
            'onchange',
            [
                {
                    'selector': '#div_id_field',
                    'values': {
                        SingleValueReport.SINGLE_VALUE_TYPE_COUNT: 'hide',
                        SingleValueReport.SINGLE_VALUE_TYPE_PERCENT_FROM_COUNT: 'hide',
                    },
                    'default': 'show',
                },
                {
                    'selector': '#div_id_prefix',
                    'values': {
                        SingleValueReport.SINGLE_VALUE_TYPE_COUNT: 'hide',
                        SingleValueReport.SINGLE_VALUE_TYPE_PERCENT_FROM_COUNT: 'hide',
                    },
                    'default': 'show',
                },
                {
                    'selector': '#div_id_numerator',
                    'values': {SingleValueReport.SINGLE_VALUE_TYPE_PERCENT: 'show'},
                    'default': 'hide',
                },
                {
                    'selector': 'label[for=id_field]',
                    'values': {
                        SingleValueReport.SINGLE_VALUE_TYPE_PERCENT: (
                            'html',
                            'Denominator field',
                        )
                    },
                    'default': ('html', 'Field'),
                },
                {
                    'selector': '#div_id_average_scale',
                    'values': {SingleValueReport.SINGLE_VALUE_TYPE_AVERAGE_SUM_OVER_TIME: 'show'},
                    'default': 'hide',
                },
                {
                    'selector': '#div_id_average_start_period',
                    'values': {SingleValueReport.SINGLE_VALUE_TYPE_AVERAGE_SUM_OVER_TIME: 'show'},
                    'default': 'hide',
                },
                {
                    'selector': '#div_id_average_end_period',
                    'values': {SingleValueReport.SINGLE_VALUE_TYPE_AVERAGE_SUM_OVER_TIME: 'show'},
                    'default': 'hide',
                },
            ],
        )

        form.add_trigger(
            'show_breakdown',
            'onchange',
            [
                {
                    'selector': '#div_id_breakdown_fields',
                    'values': {'checked': 'show'},
                    'default': 'hide',
                },
            ],
        )

        if 'data' in _kwargs and len(_kwargs['data']) > 0:
            field = _kwargs['data'].get('field')
            numerator = _kwargs['data'].get('numerator')
            report_type_id = _kwargs['data'].get('report_type')
            report_type = get_object_or_404(ReportType, id=report_type_id)
        else:
            field = form.instance.field
            report_type = form.instance.report_type
            numerator = form.instance.numerator

        self.setup_field(
            field_type='number',
            form=form,
            field_name='field',
            selected_field_id=field,
            report_type=report_type,
        )

        self.setup_field(
            field_type='number',
            form=form,
            field_name='numerator',
            selected_field_id=numerator,
            report_type=report_type,
        )

        form.fields['notes'].widget.attrs['rows'] = 3
        url = reverse(
            'advanced_report_builder:single_value_field_modal',
            kwargs={'slug': 'selector-99999-data-FIELD_INFO-report_type_id-REPORT_TYPE_ID'},
        )

        range_type_choices = VariableDate.RANGE_TYPE_CHOICES
        form.fields['average_start_period'] = ChoiceField(required=False, choices=range_type_choices)
        form.fields['average_end_period'] = ChoiceField(required=False, choices=range_type_choices)
        fields = [
            'name',
            'notes',
            'report_type',
            'report_tags',
            'single_value_type',
            'numerator',
            'field',
            'average_scale',
            'average_start_period',
            'average_end_period',
            'prefix',
            'tile_colour',
            'decimal_places',
            'show_breakdown',
            FieldEx(
                'breakdown_fields',
                template='advanced_report_builder/select_column.html',
                extra_context={'select_column_url': url, 'command_prefix': ''},
            ),
        ]
        if self.object.id:
            self.add_extra_queries(form=form, fields=fields)
            if self.object.single_value_type in [
                SingleValueReport.SINGLE_VALUE_TYPE_PERCENT,
                SingleValueReport.SINGLE_VALUE_TYPE_PERCENT_FROM_COUNT,
            ]:
                self.add_page_command('set_css', selector='.btn-query-edit', prop='display', val='none')
            else:
                self.add_page_command(
                    'set_css',
                    selector='.btn-query-numerator-edit',
                    prop='display',
                    val='none',
                )
        return fields

    def select2_field(self, **kwargs):
        return self.get_fields_for_select2(
            field_type='number',
            report_type=kwargs['report_type'],
            search_string=kwargs.get('search'),
        )

    def select2_numerator(self, **kwargs):
        return self.get_fields_for_select2(
            field_type='number',
            report_type=kwargs['report_type'],
            search_string=kwargs.get('search'),
        )

    # noinspection PyUnusedLocal
    def clean(self, form, cleaned_data):
        single_value_type = cleaned_data['single_value_type']
        if (
            single_value_type
            not in [
                SingleValueReport.SINGLE_VALUE_TYPE_COUNT,
                SingleValueReport.SINGLE_VALUE_TYPE_PERCENT_FROM_COUNT,
            ]
            and not cleaned_data['field']
        ):
            raise ValidationError('Please select a field')

        if single_value_type == SingleValueReport.SINGLE_VALUE_TYPE_AVERAGE_SUM_OVER_TIME:
            if cleaned_data['average_scale'] is None:
                form.add_error('average_scale', 'Please select a time scale')
            if cleaned_data['average_start_period'] is None:
                form.add_error('average_start_period', 'Please select a start period')
            if cleaned_data['average_end_period'] is None:
                form.add_error('average_end_period', 'Please select a end period')

    def ajax_get_fields(self, **kwargs):
        report_type_id = kwargs['report_type']
        report_builder_fields, base_model = self.get_report_builder_class(report_type_id=report_type_id)
        fields = []
        tables = []
        self._get_fields(
            base_model=base_model,
            fields=fields,
            tables=tables,
            report_builder_class=report_builder_fields,
        )
        return self.command_response('report_fields', data=json.dumps({'fields': fields, 'tables': tables}))

    def button_add_query(self, **_kwargs):
        single_value_type = _kwargs['single_value_type']
        if single_value_type == '' or int(single_value_type) not in [
            SingleValueReport.SINGLE_VALUE_TYPE_COUNT,
            SingleValueReport.SINGLE_VALUE_TYPE_PERCENT_FROM_COUNT,
        ]:
            return super().button_add_query(**_kwargs)
        else:
            report_type = self.get_report_type(**_kwargs)
            show_order_by = 1 if self.show_order_by else 0
            url = reverse(
                'advanced_report_builder:single_value_numerator_modal',
                kwargs={'slug': f'report_id-{self.object.id}-report_type-{report_type}-show_order_by-{show_order_by}'},
            )
            return self.command_response('show_modal', modal=url)

    def button_edit_query(self, **_kwargs):
        single_value_type = _kwargs['single_value_type']
        if single_value_type == '' or int(single_value_type) not in [
            SingleValueReport.SINGLE_VALUE_TYPE_COUNT,
            SingleValueReport.SINGLE_VALUE_TYPE_PERCENT_FROM_COUNT,
        ]:
            return super().button_edit_query(**_kwargs)
        else:
            query_id = _kwargs['query_id'][1:]
            report_type = self.get_report_type(**_kwargs)
            url = reverse(
                'advanced_report_builder:single_value_numerator_modal',
                kwargs={'slug': f'pk-{query_id}-report_type-{report_type}'},
            )
            return self.command_response('show_modal', modal=url)

    def form_valid(self, form):
        instance = form.save(commit=False)
        instance._current_user = self.request.user
        instance.save()
        return self.command_response('reload')


class SingleValueShowBreakdownModal(TableUtilsMixin, Modal):
    button_container_class = 'text-center'
    size = 'xl'

    def modal_title(self):
        return self.table_report.name

    def add_table(self, base_model):
        return DatatableTable(view=self, model=base_model)

    def setup_table(self):
        single_value_report = get_object_or_404(SingleValueReport, pk=self.slug['pk'])
        self.kwargs['enable_links'] = self.slug['enable_links'] == 'True'
        self.table_report = single_value_report
        base_model = single_value_report.get_base_model()
        table = self.add_table(base_model=base_model)
        table.extra_filters = self.extra_filters
        table_fields = single_value_report.breakdown_fields
        report_builder_class = get_report_builder_class(model=base_model, report_type=self.table_report.report_type)
        fields_used = set()
        fields_map = {}
        self.process_query_results(
            report_builder_class=report_builder_class,
            table=table,
            base_model=base_model,
            fields_used=fields_used,
            fields_map=fields_map,
            table_fields=table_fields,
        )

        table.ajax_data = False
        table.table_options['pageLength'] = 25
        table.table_options['bStateSave'] = False
        return table

    def modal_content(self):
        table = self.setup_table()
        return table.render()


class SingleValueTableFieldForm(TableFieldForm):
    cancel_class = 'btn-secondary modal-cancel'


class SingleValueTableFieldModal(TableFieldModal):
    form_class = SingleValueTableFieldForm


class QueryNumeratorForm(QueryBuilderModelForm):
    cancel_class = 'btn-secondary modal-cancel'

    class Meta:
        model = ReportQuery
        fields = ['name', 'query', 'extra_query']

    def setup_modal(self, *args, **kwargs):
        self.fields['extra_query'].label = 'Numerator'
        super().setup_modal(*args, **kwargs)


class QueryNumeratorModal(QueryBuilderModalBaseMixin, ModelFormModal):
    model = ReportQuery
    process = PROCESS_EDIT_DELETE
    permission_delete = PERMISSION_OFF
    form_class = QueryNumeratorForm
    helper_class = HorizontalNoEnterHelper
    ajax_commands = ['datatable', 'button']

    template_name = 'advanced_report_builder/extra_query_modal.html'
    no_header_x = True

    def __init__(self, *args, **kwargs):
        self._base_and_class = None
        super().__init__(*args, **kwargs)

    def post_save(self, created, form):
        self.add_command({'function': 'save_query_builder_id_extra_query'})
        return self.command_response('close')

    def form_setup(self, form, *_args, **_kwargs):
        fields = [
            'name',
            FieldEx('query', template='advanced_report_builder/query_builder.html'),
            FieldEx('extra_query', template='advanced_report_builder/query_builder.html'),
        ]

        return fields

    def ajax_get_query_builder_fields(self, **kwargs):
        field_auto_id = kwargs['field_auto_id']
        report_type_id = self.slug['report_type']
        query_builder_filters = self.get_query_builder_report_type_field(report_type_id=report_type_id)
        return self.command_response(f'query_builder_{field_auto_id}', data=json.dumps(query_builder_filters))

    def get_report_builder_base_and_class(self):
        if self._base_and_class is None:
            self._base_and_class = self.get_report_builder_class(report_type_id=self.slug['report_type'])
        return self._base_and_class

    def get_report_type(self):
        return self.slug['report_type']

    def form_valid(self, form):
        org_id = self.object.pk if hasattr(self, 'object') else None
        instance = form.save(commit=False)
        instance._current_user = self.request.user
        instance.save()
        self.post_save(created=org_id is None, form=form)
        if not self.response_commands:
            self.add_command('reload')
        return self.command_response()



import json

from crispy_forms.layout import HTML, Div
from django.forms import CharField
from django.urls import reverse
from django_datatables.columns import ColumnBase, MenuColumn
from django_datatables.widgets import DataTableReorderWidget
from django_menus.menu import HtmlMenu, MenuItem
from django_modals.fields import FieldEx
from django_modals.form_helpers import HorizontalNoEnterHelper
from django_modals.forms import ModelCrispyForm
from django_modals.modals import ModelFormModal
from django_modals.processes import PERMISSION_OFF, PROCESS_EDIT_DELETE

from advanced_report_builder.models import ReportQuery, ReportQueryOrder, ReportType
from advanced_report_builder.toggle import RBToggle
from advanced_report_builder.utils import get_query_js
from advanced_report_builder.views.helpers import QueryBuilderModelForm
from advanced_report_builder.views.modals_base import QueryBuilderModalBaseMixin


class QueryForm(QueryBuilderModelForm):
    cancel_class = 'btn-secondary modal-cancel'

    class Meta:
        model = ReportQuery
        fields = ['name', 'query']


class QueryModal(QueryBuilderModalBaseMixin, ModelFormModal):
    model = ReportQuery
    process = PROCESS_EDIT_DELETE
    permission_delete = PERMISSION_OFF
    form_class = QueryForm
    helper_class = HorizontalNoEnterHelper
    ajax_commands = ['datatable', 'button']

    template_name = 'advanced_report_builder/query_modal.html'
    no_header_x = True

    def __init__(self, *args, **kwargs):
        self._base_and_class = None
        super().__init__(*args, **kwargs)

    def post_save(self, created, form):
        self.add_command({'function': 'save_query_builder_id_query'})
        return self.command_response('close')

    def form_setup(self, form, *_args, **_kwargs):
        fields = [
            'name',
            FieldEx('query', template='advanced_report_builder/query_builder.html'),
        ]

        if self.object.id and self.slug.get('show_order_by') == '1':
            self.add_query_orders(form=form, fields=fields)

        return fields

    def ajax_get_query_builder_fields(self, **kwargs):
        field_auto_id = kwargs['field_auto_id']
        report_type_id = self.slug['report_type']
        query_builder_filters = self.get_query_builder_report_type_field(report_type_id=report_type_id)
        return self.command_response(f'query_builder_{field_auto_id}', data=json.dumps(query_builder_filters))

    def get_report_builder_base_and_class(self):
        if self._base_and_class is None:
            self._base_and_class = self.get_report_builder_class(report_type_id=self.slug['report_type'])
        return self._base_and_class

    def order_by_result(self, _column, data_dict, _page_results):
        order_by_field = data_dict['order_by_field']
        field_values = []
        report_builder_class, base_model = self.get_report_builder_base_and_class()
        self.get_field_display_value(
            field_type='order',
            fields_values=field_values,
            base_model=base_model,
            report_builder_class=report_builder_class,
            selected_field_value=order_by_field,
        )

        if len(field_values) > 0:
            return field_values[0]['label']

        return order_by_field

    def add_query_orders(self, form, fields):
        add_query_js = (
            'django_modal.process_commands_lock([{"function": "post_modal", "button": {"button": "add_query_order"}}])'
        )

        description_add_menu_items = [
            MenuItem(
                add_query_js.replace('"', '&quot;'),
                menu_display='Add Order by',
                css_classes='btn btn-primary',
                font_awesome='fas fa-pencil',
                link_type=MenuItem.HREF,
            )
        ]

        menu = HtmlMenu(self.request, 'advanced_report_builder/datatables/onclick_menu.html').add_items(
            *description_add_menu_items
        )

        fields.append(Div(HTML(menu.render()), css_class='form-buttons'))
        edit_query_js = get_query_js('edit_query_order', 'query_order_id')
        duplicate_query_js = get_query_js('duplicate_query_order', 'query_order_id')
        description_edit_menu_items = [
            MenuItem(
                duplicate_query_js.replace('"', '&quot;'),
                menu_display='Duplicate',
                css_classes='btn btn-sm btn-outline-dark',
                font_awesome='fas fa-clone',
                link_type=MenuItem.HREF,
            ),
            MenuItem(
                edit_query_js.replace('"', '&quot;'),
                menu_display='Edit',
                css_classes='btn btn-sm btn-outline-dark btn-query-edit',
                font_awesome='fas fa-pencil',
                link_type=MenuItem.HREF,
            ),
        ]

        template = 'advanced_report_builder/datatables/onclick_menu.html'
        form.fields['orders'] = CharField(
            required=False,
            label='Query Order',
            widget=DataTableReorderWidget(
                model=ReportQueryOrder,
                order_field='order',
                fields=[
                    '_.index',
                    '.id',
                    ColumnBase(
                        column_name='order_by',
                        field='order_by_field',
                        row_result=self.order_by_result,
                    ),
                    'order_by_ascending',
                    MenuColumn(
                        column_name='menu',
                        field='id',
                        column_defs={'orderable': False, 'className': 'dt-right'},
                        menu=HtmlMenu(self.request, template).add_items(*description_edit_menu_items),
                    ),
                ],
                attrs={'filter': {'report_query__id': self.object.id}},
            ),
        )
        fields.append('orders')

    def get_report_type(self):
        return self.slug['report_type']

    def button_duplicate_query_order(self, **_kwargs):
        query_id = _kwargs['query_order_id'][1:]
        report_query_order = ReportQueryOrder.objects.get(pk=query_id)
        report_query_order.slug = None
        report_query_order.pk = None
        report_query_order.order = None
        report_query_order.save()

        report_type = self.get_report_type()
        url = reverse(
            'advanced_report_builder:query_order_modal',
            kwargs={'slug': f'pk-{report_query_order.id}-report_type-{report_type}'},
        )
        return self.command_response('show_modal', modal=url)

    def button_add_query_order(self, **_kwargs):
        report_type = self.get_report_type()
        url = reverse(
            'advanced_report_builder:query_order_modal',
            kwargs={'slug': f'report_query_id-{self.object.id}-report_type-{report_type}'},
        )
        return self.command_response('show_modal', modal=url)

    def button_edit_query_order(self, **_kwargs):
        query_order_id = _kwargs['query_order_id'][1:]
        report_type = self.get_report_type()
        url = reverse(
            'advanced_report_builder:query_order_modal',
            kwargs={'slug': f'pk-{query_order_id}-report_type-{report_type}'},
        )
        return self.command_response('show_modal', modal=url)

    def datatable_sort(self, **kwargs):
        current_sort = dict(ReportQueryOrder.objects.filter(report_query=self.object.id).values_list('id', 'order'))
        for s in kwargs['sort']:
            if current_sort[s[1]] != s[0]:
                o = ReportQueryOrder.objects.get(id=s[1])
                o.order = s[0]
                o.save()
        return self.command_response('')

    def form_valid(self, form):
        org_id = self.object.pk if hasattr(self, 'object') else None
        instance = form.save(commit=False)
        instance._current_user = self.request.user
        instance.save()
        self.post_save(created=org_id is None, form=form)
        if not self.response_commands:
            self.add_command('reload')
        return self.command_response()


class OrderByFieldForm(ModelCrispyForm):
    class Meta:
        model = ReportQueryOrder
        fields = ['order_by_field', 'order_by_ascending']

    cancel_class = 'btn-secondary modal-cancel'

    def cancel_button(self, css_class=cancel_class, **kwargs):
        commands = [{'function': 'save_query_builder_id_query'}, {'function': 'close'}]
        return self.button('Cancel', commands, css_class, **kwargs)


class QueryOrderModal(QueryBuilderModalBaseMixin, ModelFormModal):
    form_class = OrderByFieldForm
    ajax_commands = ['datatable', 'button']
    model = ReportQueryOrder
    process = PROCESS_EDIT_DELETE
    permission_delete = PERMISSION_OFF
    helper_class = HorizontalNoEnterHelper
    no_header_x = True

    def form_setup(self, form, *_args, **_kwargs):
        form.fields['order_by_ascending'].widget = RBToggle()

        report_type = ReportType.objects.get(id=self.slug['report_type'])
        order_by_field = _kwargs['data'].get('order_by_field') if 'data' in _kwargs else form.instance.order_by_field

        self.setup_field(
            field_type='order',
            form=form,
            field_name='order_by_field',
            selected_field_id=order_by_field,
            report_type=report_type,
        )

    def select2_order_by_field(self, **kwargs):
        return self.get_fields_for_select2(
            field_type='django_order',
            report_type=self.slug['report_type'],
            search_string=kwargs.get('search'),
        )

    def post_save(self, created, form):
        self.add_command({'function': 'save_query_builder_id_query'})
        return self.command_response('close')



from django.contrib import admin

from advanced_report_builder.models import (
    BarChartReport,
    CustomReport,
    Dashboard,
    DashboardReport,
    FunnelChartReport,
    KanbanReport,
    KanbanReportDescription,
    KanbanReportLane,
    LineChartReport,
    PieChartReport,
    Report,
    ReportQuery,
    ReportTag,
    ReportType,
    SingleValueReport,
    TableReport,
    Target,
)


@admin.register(ReportQuery)
class ReportQueryAdmin(admin.ModelAdmin):
    list_display = ('report', 'name', 'query')


class ReportQueryInline(admin.TabularInline):
    model = ReportQuery


@admin.register(Report)
class ReportAdmin(admin.ModelAdmin):
    list_display = ('name', 'instance_type', 'slug', 'slug_alias', 'version')
    exclude = ('slug',)

    inlines = [ReportQueryInline]


@admin.register(ReportType)
class ReportTypeAdmin(admin.ModelAdmin):
    list_display = ('name', 'content_type', 'report_builder_class_name')


@admin.register(TableReport)
class TableReportAdmin(admin.ModelAdmin):
    list_display = (
        'name',
        'slug',
    )
    exclude = (
        'instance_type',
        'slug',
    )
    inlines = [ReportQueryInline]


@admin.register(SingleValueReport)
class SingleValueReportAdmin(admin.ModelAdmin):
    list_display = (
        'name',
        'slug',
    )
    exclude = (
        'instance_type',
        'slug',
    )
    inlines = [ReportQueryInline]


@admin.register(BarChartReport)
class BarChartReportAdmin(admin.ModelAdmin):
    list_display = (
        'name',
        'slug',
    )
    exclude = (
        'instance_type',
        'slug',
    )
    inlines = [ReportQueryInline]


@admin.register(LineChartReport)
class LineChartReportAdmin(admin.ModelAdmin):
    list_display = (
        'name',
        'slug',
    )
    exclude = (
        'instance_type',
        'slug',
    )
    inlines = [ReportQueryInline]


@admin.register(PieChartReport)
class PieChartReportAdmin(admin.ModelAdmin):
    list_display = (
        'name',
        'slug',
    )
    exclude = (
        'instance_type',
        'slug',
    )
    inlines = [ReportQueryInline]


@admin.register(FunnelChartReport)
class FunnelChartReportAdmin(admin.ModelAdmin):
    list_display = (
        'name',
        'slug',
    )
    exclude = (
        'instance_type',
        'slug',
    )
    inlines = [ReportQueryInline]


@admin.register(CustomReport)
class CustomReportAdmin(admin.ModelAdmin):
    list_display = ('name', 'slug', 'output_type', 'view_name', 'settings')
    exclude = ('instance_type', 'slug', 'report_type')


class KanbanReportDescriptionInline(admin.TabularInline):
    model = KanbanReportDescription


class KanbanReportLaneInline(admin.TabularInline):
    model = KanbanReportLane


@admin.register(KanbanReport)
class KanbanReportAdmin(admin.ModelAdmin):
    list_display = (
        'name',
        'slug',
    )
    exclude = ('instance_type', 'slug', 'report_type')
    inlines = [KanbanReportLaneInline, KanbanReportDescriptionInline]


class DashboardReportInline(admin.TabularInline):
    model = DashboardReport


@admin.register(Dashboard)
class DashboardAdmin(admin.ModelAdmin):
    list_display = (
        'slug',
        'name',
    )

    exclude = [
        'slug',
    ]

    inlines = [DashboardReportInline]


@admin.register(ReportTag)
class ReportTagAdmin(admin.ModelAdmin):
    list_display = ('name', 'order')


@admin.register(Target)
class TargetAdmin(admin.ModelAdmin):
    list_display = ('pk', 'name', 'target_type', 'default_value')



from crispy_forms.bootstrap import StrictButton
from django_modals.forms import CrispyForm, ModelCrispyForm


class QueryBuilderModelForm(ModelCrispyForm):
    def submit_button(self, css_class='btn-success modal-submit', button_text='Submit', **kwargs):
        return StrictButton(
            button_text,
            onclick=f'save_modal_{self.form_id}()',
            css_class=css_class,
            **kwargs,
        )


class QueryBuilderForm(CrispyForm):
    def submit_button(self, css_class='btn-success modal-submit', button_text='Submit', **kwargs):
        return StrictButton(
            button_text,
            onclick=f'save_modal_{self.form_id}()',
            css_class=css_class,
            **kwargs,
        )



from crispy_forms.layout import HTML, Div
from django.conf import settings
from django.forms import CharField
from django.urls import reverse
from django_datatables.columns import MenuColumn
from django_datatables.widgets import DataTableReorderWidget
from django_menus.menu import HtmlMenu, MenuItem

from advanced_report_builder.models import ReportQuery
from advanced_report_builder.utils import get_query_js


class MultiQueryModalMixin:
    show_order_by = True

    ajax_commands = ['datatable', 'button', 'ajax']

    def datatable_sort(self, **kwargs):
        form = self.get_form()
        widget = form.fields[kwargs['table_id'][3:]].widget
        _model = widget.attrs['table_model']
        current_sort = dict(_model.objects.filter(report_id=self.object.id).values_list('id', 'order'))
        for s in kwargs['sort']:
            if current_sort[s[1]] != s[0]:
                o = _model.objects.get(id=s[1])
                o.order = s[0]
                o.save()
        return self.command_response('null')

    def get_report_type(self, **_kwargs):
        return _kwargs['report_type']

    def button_duplicate_query(self, **_kwargs):
        query_id = _kwargs['query_id'][1:]
        report_query = ReportQuery.objects.get(pk=query_id)
        report_query_orders = report_query.reportqueryorder_set.all()
        report_query.slug = None
        report_query.pk = None
        report_query.order = None
        report_query.save()

        for report_query_order in report_query_orders:
            report_query_order.pk = None
            report_query_order.report_query = report_query
            report_query_order.save()

        report_type = self.get_report_type(**_kwargs)
        url = reverse(
            'advanced_report_builder:query_modal',
            kwargs={'slug': f'pk-{report_query.id}-report_type-{report_type}'},
        )
        return self.command_response('show_modal', modal=url)

    def button_add_query(self, **_kwargs):
        report_type = self.get_report_type(**_kwargs)
        show_order_by = 1 if self.show_order_by else 0
        url = reverse(
            'advanced_report_builder:query_modal',
            kwargs={'slug': f'report_id-{self.object.id}-report_type-{report_type}-show_order_by-{show_order_by}'},
        )
        return self.command_response('show_modal', modal=url)

    def button_edit_query(self, **_kwargs):
        query_id = _kwargs['query_id'][1:]
        report_type = self.get_report_type(**_kwargs)
        show_order_by = 1 if self.show_order_by else 0
        url = reverse(
            'advanced_report_builder:query_modal',
            kwargs={'slug': f'pk-{query_id}-report_type-{report_type}-show_order_by-{show_order_by}'},
        )
        return self.command_response('show_modal', modal=url)

    def query_menu(self):
        edit_query_js = get_query_js('edit_query', 'query_id')
        duplicate_query_js = get_query_js('duplicate_query', 'query_id')

        description_edit_menu_items = [
            MenuItem(
                duplicate_query_js.replace('"', '&quot;'),
                menu_display='Duplicate',
                css_classes='btn btn-sm btn-outline-dark',
                font_awesome='fas fa-clone',
                link_type=MenuItem.HREF,
            ),
            MenuItem(
                edit_query_js.replace('"', '&quot;'),
                menu_display='Edit',
                css_classes='btn btn-sm btn-outline-dark',
                font_awesome='fas fa-pencil',
                link_type=MenuItem.HREF,
            ),
        ]
        return description_edit_menu_items

    def add_extra_queries(self, form, fields):
        add_query_js = (
            'django_modal.process_commands_lock([{"function": "post_modal", "button": {"button": "add_query"}}])'
        )

        description_add_menu_items = [
            MenuItem(
                add_query_js.replace('"', '&quot;'),
                menu_display='Add Query Version',
                css_classes='btn btn-primary',
                font_awesome='fas fa-pencil',
                link_type=MenuItem.HREF,
            )
        ]

        menu = HtmlMenu(self.request, 'advanced_report_builder/datatables/onclick_menu.html').add_items(
            *description_add_menu_items
        )

        fields.append(Div(HTML(menu.render()), css_class='form-buttons'))
        description_edit_menu_items = self.query_menu()

        form.fields['queries'] = CharField(
            required=False,
            label='Queries Versions',
            widget=DataTableReorderWidget(
                model=ReportQuery,
                order_field='order',
                fields=[
                    '_.index',
                    '.id',
                    'name',
                    MenuColumn(
                        column_name='menu',
                        field='id',
                        column_defs={'orderable': False, 'className': 'dt-right'},
                        menu=HtmlMenu(
                            self.request,
                            'advanced_report_builder/datatables/onclick_menu.html',
                        ).add_items(*description_edit_menu_items),
                    ),
                ],
                attrs={'filter': {'report__id': self.object.id}},
            ),
        )
        fields.append('queries')

    def post_save(self, created, form):
        if created:
            self.modal_redirect(
                self.request.resolver_match.view_name,
                slug=f'pk-{self.object.id}-new-True',
            )
        else:
            url_name = getattr(settings, 'REPORT_BUILDER_DETAIL_URL_NAME', '')
            if url_name and self.slug.get('new'):
                url = reverse(url_name, kwargs={'slug': self.object.slug})
                self.command_response('redirect', url=url)



from django.db import models
from django_datatables.columns import ManyToManyColumn

from advanced_report_builder.column_types import DATE_FIELDS
from advanced_report_builder.columns import FilterForeignKeyColumn
from advanced_report_builder.variable_date import VariableDate


class FieldTypes:
    FIELD_TYPE_STRING = 1
    FIELD_TYPE_NUMBER = 2
    FIELD_TYPE_DATE = 3
    FIELD_TYPE_BOOLEAN = 4
    FIELD_TYPE_MULTIPLE_CHOICE = 5
    FIELD_TYPE_FOREIGN_KEY = 6
    FIELD_TYPE_ABSTRACT_USER = 7
    FIELD_TYPE_PART_DATE = 8

    def get_operator(self, field_type):
        operators = {
            self.FIELD_TYPE_STRING: [
                'equal',
                'not_equal',
                'contains',
                'not_contains',
                'begins_with',
                'not_begins_with',
                'ends_with',
                'not_ends_with',
            ],
            self.FIELD_TYPE_NUMBER: [
                'equal',
                'not_equal',
                'less',
                'less_or_equal',
                'greater',
                'greater_or_equal',
            ],
            self.FIELD_TYPE_DATE: [
                'equal',
                'not_equal',
                'less',
                'less_or_equal',
                'greater',
                'greater_or_equal',
                'is_null',
                'is_not_null',
            ],
            self.FIELD_TYPE_PART_DATE: ['equal', 'not_equal', 'is_null', 'is_not_null'],
            self.FIELD_TYPE_BOOLEAN: ['equal', 'not_equal'],
            self.FIELD_TYPE_MULTIPLE_CHOICE: ['in', 'not_in'],
            self.FIELD_TYPE_FOREIGN_KEY: ['is_null', 'is_not_null'],
            self.FIELD_TYPE_ABSTRACT_USER: ['equal', 'not_equal'],
        }
        return operators.get(field_type)

    def get_filter(self, query_builder_filters, django_field, field, title, column, prefix):
        if isinstance(column.field, list):
            full_field_name = prefix + column.field[0]
        elif column.field is not None:
            full_field_name = prefix + column.field
        else:
            full_field_name = prefix + field
        column_id = prefix + field
        if django_field is not None and not isinstance(column, ManyToManyColumn):
            if isinstance(column, FilterForeignKeyColumn):
                query_builder_filters.append(
                    {
                        'id': column_id,
                        'label': title,
                        'field': full_field_name,
                        'type': 'string',
                        'input': 'select',
                        'multiple': True,
                        'values': column.get_query_options(),
                        'operators': self.get_operator(self.FIELD_TYPE_MULTIPLE_CHOICE),
                    }
                )
            elif isinstance(django_field, models.CharField | models.TextField | models.EmailField):
                query_builder_filters.append(
                    {
                        'id': column_id,
                        'label': title,
                        'field': full_field_name,
                        'type': 'string',
                        'operators': self.get_operator(self.FIELD_TYPE_STRING),
                        'validation': {'allow_empty_value': True},
                    }
                )
            elif isinstance(
                django_field,
                models.IntegerField | models.PositiveSmallIntegerField | models.PositiveIntegerField,
            ):
                if django_field.choices is None:
                    query_builder_filter = {
                        'id': column_id,
                        'label': title,
                        'field': full_field_name,
                        'type': 'integer',
                        'operators': self.get_operator(self.FIELD_TYPE_NUMBER),
                    }
                else:
                    query_builder_filter = {
                        'id': column_id,
                        'label': title,
                        'field': full_field_name,
                        'type': 'integer',
                        'input': 'select',
                        'multiple': True,
                        'values': dict(django_field.choices),
                        'operators': self.get_operator(self.FIELD_TYPE_MULTIPLE_CHOICE),
                    }
                query_builder_filters.append(query_builder_filter)
            elif isinstance(django_field, models.BooleanField):
                query_builder_filters.append(
                    {
                        'id': column_id,
                        'label': title,
                        'field': full_field_name,
                        'input': 'select',
                        'operators': self.get_operator(self.FIELD_TYPE_BOOLEAN),
                        'values': {'0': 'False', '1': 'True'},
                    }
                )
            elif isinstance(django_field, DATE_FIELDS):
                self.get_date_field(
                    column_id=column_id,
                    query_builder_filters=query_builder_filters,
                    field=full_field_name,
                    title=title,
                )
        elif isinstance(column, ManyToManyColumn):
            choices = dict(column.options['lookup'])
            query_builder_filter = {
                'id': column_id,
                'label': title,
                'field': prefix + column.field_id,
                'type': 'integer',
                'input': 'select',
                'multiple': True,
                'values': choices,
                'operators': self.get_operator(self.FIELD_TYPE_MULTIPLE_CHOICE),
            }
            query_builder_filters.append(query_builder_filter)

    def get_foreign_key_null_field(self, query_builder_filters, field, title):
        query_builder_filters.append(
            {
                'id': field,
                'label': title,
                'field': field,
                'type': 'integer',
                'operators': self.get_operator(self.FIELD_TYPE_FOREIGN_KEY),
                'validation': {'allow_empty_value': True},
            }
        )

    def get_date_field(self, column_id, query_builder_filters, field, title):
        variable_date = VariableDate()
        query_builder_filter = {
            'id': f'{column_id}__variable_date',
            'label': f'{title} (Variable)',
            'field': field,
            'operators': self.get_operator(self.FIELD_TYPE_DATE),
            'input': 'select',
            'values': variable_date.get_variable_date_filter_values(),
        }
        query_builder_filters.append(query_builder_filter)
        query_builder_filter = {
            'id': f'{column_id}__variable_year',
            'label': f'{title} (Year)',
            'field': field,
            'operators': self.get_operator(self.FIELD_TYPE_DATE),
            'input': 'select',
            'values': variable_date.get_date_filter_years(),
        }
        query_builder_filters.append(query_builder_filter)
        query_builder_filter = {
            'id': f'{column_id}__variable_month',
            'label': f'{title} (Month)',
            'field': field,
            'operators': self.get_operator(self.FIELD_TYPE_DATE),
            'input': 'select',
            'values': variable_date.get_date_filter_months(),
        }
        query_builder_filters.append(query_builder_filter)

        query_builder_filter = {
            'id': f'{column_id}__variable_quarter',
            'label': f'{title} (Quarter)',
            'field': field,
            'operators': self.get_operator(self.FIELD_TYPE_PART_DATE),
            'input': 'select',
            'values': variable_date.get_date_filter_quarters(),
        }
        query_builder_filters.append(query_builder_filter)

    def get_abstract_user_field(self, query_builder_filters, field, title):
        query_builder_filters.append(
            {
                'id': f'{field}__logged_in_user',
                'label': f'{title} (Logged in user)',
                'field': field,
                'input': 'select',
                'operators': self.get_operator(self.FIELD_TYPE_ABSTRACT_USER),
                'values': {'0': 'False', '1': 'True'},
            }
        )



from django.apps import AppConfig


class AdvancedReportBuilderConfig(AppConfig):
    default_auto_field = 'django.db.models.AutoField'
    name = 'advanced_report_builder'
    verbose_name = 'Advanced Report builder'



# Generated by Django 3.2.7 on 2023-11-08 09:26

from django.db import migrations, models
import django.db.models.deletion
import time_stamped_model.models


class Migration(migrations.Migration):

    dependencies = [
        ('advanced_report_builder', '0012_auto_20230718_1642'),
    ]

    operations = [
        migrations.CreateModel(
            name='ReportQueryOrder',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', time_stamped_model.models.CreationDateTimeField(auto_now_add=True)),
                ('modified', time_stamped_model.models.ModificationDateTimeField(auto_now=True)),
                ('order_by_field', models.CharField(blank=True, max_length=200, null=True)),
                ('order_by_ascending', models.BooleanField(default=True)),
                ('order', models.PositiveSmallIntegerField()),
                ('report_query', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='advanced_report_builder.reportquery')),
            ],
            options={
                'ordering': ['order'],
            },
        ),
    ]



from django.db import models
from django_datatables.columns import ColumnLink, CurrencyColumn, CurrencyPenceColumn

from advanced_report_builder.columns import (
    ColourColumn,
    ReverseForeignKeyBoolColumn,
    ReverseForeignKeyChoiceColumn,
    ReverseForeignKeyDateColumn,
    ReverseForeignKeyStrColumn,
)

NUMBER_FIELDS = (
    models.IntegerField,
    models.PositiveSmallIntegerField,
    models.PositiveIntegerField,
    models.FloatField,
)
DATE_FIELDS = (models.DateTimeField, models.DateField)
BOOLEAN_FIELD = models.BooleanField
CURRENCY_COLUMNS = (CurrencyColumn, CurrencyPenceColumn)
LINK_COLUMNS = ColumnLink
COLOUR_COLUMNS = ColourColumn
REVERSE_FOREIGN_KEY_STR_COLUMNS = ReverseForeignKeyStrColumn
REVERSE_FOREIGN_KEY_BOOL_COLUMNS = ReverseForeignKeyBoolColumn
REVERSE_FOREIGN_KEY_CHOICE_COLUMNS = ReverseForeignKeyChoiceColumn
REVERSE_FOREIGN_KEY_DATE_COLUMNS = ReverseForeignKeyDateColumn



# Generated by Django 3.2.7 on 2024-07-06 22:07

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('advanced_report_builder', '0013_reportqueryorder'),
    ]

    operations = [
        migrations.AddField(
            model_name='dashboardreport',
            name='report_query',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, to='advanced_report_builder.reportquery'),
        ),
        migrations.AddField(
            model_name='dashboardreport',
            name='show_versions',
            field=models.BooleanField(default=True),
        ),
    ]



import json

from django.apps import apps
from django.conf import settings
from django.contrib.auth.models import AbstractUser
from django.forms import JSONField
from django.http import JsonResponse
from django.shortcuts import get_object_or_404
from django.urls import reverse
from django_modals.modals import ModelFormModal
from django_modals.widgets.select2 import Select2

from advanced_report_builder.column_types import NUMBER_FIELDS
from advanced_report_builder.field_types import FieldTypes
from advanced_report_builder.field_utils import ReportBuilderFieldUtils
from advanced_report_builder.models import ReportQuery, ReportType
from advanced_report_builder.utils import get_report_builder_class


class QueryBuilderModalBaseMixin(ReportBuilderFieldUtils):
    @staticmethod
    def get_report_builder_class(report_type_id):
        if not report_type_id:
            return None, None

        report_type = get_object_or_404(ReportType, pk=report_type_id)
        base_model = report_type.content_type.model_class()
        report_builder_class = get_report_builder_class(model=base_model, report_type=report_type)
        return report_builder_class, base_model

    def get_query_builder_report_type_field(self, report_type_id):
        report_builder_class, base_model = self.get_report_builder_class(report_type_id=report_type_id)
        if report_builder_class is None:
            # noinspection PyUnresolvedReferences
            return self.command_response()
        query_builder_filters = []
        self._get_query_builder_fields(
            base_model=base_model,
            query_builder_filters=query_builder_filters,
            report_builder_class=report_builder_class,
        )

        return query_builder_filters

    def _get_query_builder_fields(
        self,
        base_model,
        query_builder_filters,
        report_builder_class,
        prefix='',
        title_prefix='',
        previous_base_model=None,
    ):
        field_types = FieldTypes()

        for report_builder_field in report_builder_class.fields:
            if (
                not isinstance(report_builder_field, str)
                or report_builder_field not in report_builder_class.exclude_search_fields
            ):
                django_field, _, columns, _ = self.get_field_details(
                    base_model=base_model,
                    field=report_builder_field,
                    report_builder_class=report_builder_class,
                )
                for column in columns:
                    field_types.get_filter(
                        query_builder_filters=query_builder_filters,
                        django_field=django_field,
                        field=column.column_name,
                        title=title_prefix + column.title,
                        column=column,
                        prefix=prefix,
                    )
        for include_field, include in report_builder_class.includes.items():
            app_label, model, report_builder_fields_str = include['model'].split('.')
            new_model = apps.get_model(app_label, model)
            new_report_builder_class = get_report_builder_class(model=new_model, class_name=report_builder_fields_str)

            if new_model != previous_base_model:
                _foreign_key = getattr(base_model, include_field, None)

                add_null_field = _foreign_key.field.null if hasattr(_foreign_key, 'field') else True

                if add_null_field:
                    field_types.get_foreign_key_null_field(
                        query_builder_filters=query_builder_filters,
                        field=prefix + include_field,
                        title=title_prefix + include['title'],
                    )
                if isinstance(new_model(), AbstractUser):
                    field_types.get_abstract_user_field(
                        query_builder_filters=query_builder_filters,
                        field=prefix + include_field,
                        title=title_prefix + include['title'],
                    )

                self._get_query_builder_fields(
                    base_model=new_model,
                    query_builder_filters=query_builder_filters,
                    report_builder_class=new_report_builder_class,
                    prefix=f'{prefix}{include_field}__',
                    title_prefix=f'{include["title"]} --> ',
                    previous_base_model=base_model,
                )

    def ajax_get_fields(self, **kwargs):
        report_type_id = kwargs['report_type']
        report_builder_class, base_model = self.get_report_builder_class(report_type_id=report_type_id)
        fields = []
        tables = []
        self._get_fields(
            base_model=base_model,
            fields=fields,
            tables=tables,
            report_builder_class=report_builder_class,
            field_types=NUMBER_FIELDS,
            extra_fields=report_builder_class.extra_chart_field,
        )
        return self.command_response('report_fields', data=json.dumps({'fields': fields, 'tables': tables}))

    def get_fields_for_select2(self, field_type, report_type, search_string):
        fields = []
        if report_type != '':
            report_builder_fields, base_model = self.get_report_builder_class(report_type_id=report_type)
            fields = []
            if field_type == 'date':
                self._get_date_fields(
                    base_model=base_model,
                    fields=fields,
                    report_builder_class=report_builder_fields,
                    search_string=search_string,
                )
            elif field_type == 'number':
                self._get_number_fields(
                    base_model=base_model,
                    fields=fields,
                    report_builder_class=report_builder_fields,
                    search_string=search_string,
                )
            elif field_type == 'link':
                self._get_column_link_fields(
                    base_model=base_model,
                    fields=fields,
                    report_builder_class=report_builder_fields,
                    search_string=search_string,
                )
            elif field_type == 'colour':
                self._get_colour_fields(
                    base_model=base_model,
                    fields=fields,
                    report_builder_class=report_builder_fields,
                    search_string=search_string,
                )
            elif field_type == 'all':
                self._get_fields(
                    base_model=base_model,
                    fields=fields,
                    report_builder_class=report_builder_fields,
                    for_select2=True,
                    search_string=search_string,
                )
            elif field_type in ('order', 'django_order'):
                self._get_fields(
                    base_model=base_model,
                    fields=fields,
                    report_builder_class=report_builder_fields,
                    for_select2=True,
                    search_string=search_string,
                    show_order_by_fields=True,
                    must_have_django_field=field_type == 'django_order',
                )

        return JsonResponse({'results': fields})

    def setup_field(self, field_type, form, field_name, selected_field_id, report_type):
        _fields = []
        if selected_field_id:
            form.fields[field_name].initial = selected_field_id
            base_model = report_type.content_type.model_class()
            report_builder_class = get_report_builder_class(model=base_model, report_type=report_type)
            self.get_field_display_value(
                field_type=field_type,
                fields_values=_fields,
                base_model=base_model,
                report_builder_class=report_builder_class,
                selected_field_value=selected_field_id,
                for_select2=True,
            )

        form.fields[field_name].widget = Select2(attrs={'ajax': True})
        form.fields[field_name].widget.select_data = _fields


class QueryBuilderModalBase(QueryBuilderModalBaseMixin, ModelFormModal):
    size = 'xl'

    def __init__(self, *args, **kwargs):
        self.report_query = None
        self.show_query_name = False
        super().__init__(*args, **kwargs)

    def ajax_get_query_builder_fields(self, **kwargs):
        report_type_id = kwargs['report_type']

        field_auto_id = kwargs['field_auto_id']
        if report_type_id:
            query_builder_filters = self.get_query_builder_report_type_field(report_type_id=report_type_id)

            return self.command_response(f'query_builder_{field_auto_id}', data=json.dumps(query_builder_filters))
        else:
            return self.command_response()

    def add_query_data(self, form):
        form.fields['extra_query_data'] = JSONField(required=False, label='Numerator filter')
        if self.object.id:
            query_id = self.slug.get('query_id')
            if query_id:
                self.report_query = get_object_or_404(ReportQuery, id=query_id)
            else:
                self.report_query = self.object.reportquery_set.first()
            if self.report_query:
                form.fields['extra_query_data'].initial = self.report_query.extra_query

    def form_valid(self, form):
        org_id = self.object.id if hasattr(self, 'object') else None
        chart_report = form.save(commit=False)
        chart_report._current_user = self.request.user
        chart_report.save()

        self.post_save(created=org_id is None, form=form)
        if not self.response_commands:
            url_name = getattr(settings, 'REPORT_BUILDER_DETAIL_URL_NAME', '')
            if org_id is None and url_name:
                url = reverse(url_name, kwargs={'slug': chart_report.slug})
                self.add_command('redirect', url=url)
            else:
                self.add_command('reload')
        return self.command_response()



import copy
import json

from django.db.models import ExpressionWrapper, F, FloatField, Q
from django.db.models.functions import NullIf
from django.template import Context, Template, TemplateSyntaxError
from django_datatables.columns import ColumnBase
from django_datatables.helpers import render_replace
from django_datatables.plugins.column_totals import ColumnTotals

from advanced_report_builder.column_types import (
    CURRENCY_COLUMNS,
    DATE_FIELDS,
    LINK_COLUMNS,
    NUMBER_FIELDS,
    REVERSE_FOREIGN_KEY_BOOL_COLUMNS,
    REVERSE_FOREIGN_KEY_CHOICE_COLUMNS,
    REVERSE_FOREIGN_KEY_DATE_COLUMNS,
    REVERSE_FOREIGN_KEY_STR_COLUMNS,
)
from advanced_report_builder.columns import ReportBuilderDateColumn
from advanced_report_builder.globals import (
    ALIGNMENT_CHOICE_RIGHT,
    ALIGNMENT_CLASS,
    ANNOTATION_CHOICE_NA,
    ANNOTATION_VALUE_FUNCTIONS,
    DATE_FORMAT_TYPE_DD_MM_YY_SLASH,
    DATE_FORMAT_TYPES_DJANGO_FORMAT,
    REVERSE_FOREIGN_KEY_ANNOTATION_BOOLEAN_XOR,
    REVERSE_FOREIGN_KEY_DELIMITER_COMMA,
)
from advanced_report_builder.utils import decode_attribute, split_attr
from advanced_report_builder.views.report_utils_mixin import ReportUtilsMixin


class TableUtilsMixin(ReportUtilsMixin):
    date_field = ReportBuilderDateColumn
    column_totals_class = ColumnTotals

    def __init__(self, *args, **kwargs):
        self.table_report = None
        super().__init__(*args, **kwargs)

    def get_date_field(self, index, col_type_override, table_field, fields):
        data_attr = split_attr(table_field)
        field_name = table_field['field']
        date_format = data_attr.get('date_format')

        if date_format:
            date_format = DATE_FORMAT_TYPES_DJANGO_FORMAT.get(int(date_format))

        if int(data_attr.get('display_heading', 1)) == 0:
            display_heading = False
            title = ''
        else:
            display_heading = True
            title = table_field.get('title')

        annotations_value = int(data_attr.get('annotations_value', 0))
        if col_type_override and annotations_value == 0:
            col_type_override.table = None
            field = copy.deepcopy(col_type_override)
            if title or not display_heading:
                field.title = title
            fields.append(field)
        else:
            if col_type_override:
                field_name = col_type_override.field
            date_function_kwargs = {'title': title, 'date_format': date_format}

            if annotations_value != 0:
                new_field_name = f'{annotations_value}_{field_name}_{index}'
                new_field_name = new_field_name.replace('__', '_')
                function = ANNOTATION_VALUE_FUNCTIONS[annotations_value]
                date_function_kwargs['annotations_value'] = {new_field_name: function(field_name)}
                field_name = new_field_name

            date_function_kwargs.update({'field': field_name, 'column_name': field_name})

            field = self.date_field(**date_function_kwargs)
            fields.append(field)

        return field_name

    @staticmethod
    def get_link_field(table_field, fields, col_type_override):
        field_name = table_field['field']
        col_type_override.table = None
        field = copy.deepcopy(col_type_override)
        data_attr = split_attr(table_field)
        link_css = ''
        if 'link_css' in data_attr:
            link_css = decode_attribute(data_attr['link_css'])

        link_html = ''
        if 'link_html' in data_attr:
            link_html = decode_attribute(data_attr['link_html'])

        if link_css or link_html:
            field.setup_link(link_css=link_css, link_html=link_html)

        is_icon = data_attr.get('is_icon')
        if is_icon == '1':
            field.title = ''
            field.options['no_col_search'] = True
            field.column_defs['orderable'] = False
            field.column_defs['width'] = '10px'

        elif 'title' in table_field:
            field.title = table_field['title']
        fields.append(field)

        return field_name

    def process_query_results(
        self,
        report_builder_class,
        table,
        base_model,
        fields_used,
        fields_map,
        table_fields,
        pivot_fields=None,
    ):
        first_field_name = None

        field_name = None
        fields = []
        totals = {}

        has_annotations = False
        if not table_fields:
            return fields, totals, first_field_name

        for index, table_field in enumerate(table_fields):
            field = table_field['field']
            field_attr = {}
            if 'title' in table_field:
                field_attr['title'] = table_field['title']

            original_field_name = field
            fields_used.add(field)
            django_field, col_type_override, _, _ = self.get_field_details(
                base_model=base_model,
                field=field,
                table=table,
                report_builder_class=report_builder_class,
                field_attr=field_attr,
            )

            data_attr = split_attr(table_field)
            if int(data_attr.get('display_heading', 1)) == 0:
                field_attr['title'] = ''

            annotations_type = int(data_attr.get('annotations_type', 0))
            append_annotation_query = int(data_attr.get('append_annotation_query', 0))

            if annotations_type != 0 or append_annotation_query != 0:
                has_annotations = True

            if isinstance(django_field, DATE_FIELDS):
                field_name = self.get_date_field(
                    index=index,
                    col_type_override=col_type_override,
                    table_field=table_field,
                    fields=fields,
                )
            elif isinstance(django_field, NUMBER_FIELDS) and (django_field is None or django_field.choices is None):
                decimal_places = data_attr.get('decimal_places')
                append_annotation_query = int(data_attr.get('append_annotation_query', 0)) == 1

                field_name = self.get_number_fields(
                    field_name=field_name,
                    table=table,
                    base_model=base_model,
                    report_builder_class=report_builder_class,
                    annotations_type=annotations_type,
                    append_annotation_query=append_annotation_query,
                    index=index,
                    data_attr=data_attr,
                    table_field=table_field,
                    fields=fields,
                    totals=totals,
                    col_type_override=col_type_override,
                    decimal_places=decimal_places,
                )

            elif isinstance(col_type_override, REVERSE_FOREIGN_KEY_STR_COLUMNS):
                field_name = table_field['field']
                field = self.get_reverse_foreign_key_str_field(
                    col_type_override=col_type_override,
                    data_attr=data_attr,
                    field_attr=field_attr,
                    index=index,
                )
                fields.append(field)
            elif isinstance(col_type_override, REVERSE_FOREIGN_KEY_BOOL_COLUMNS):
                field_name = table_field['field']
                field = self.get_reverse_foreign_key_bool_field(
                    col_type_override=col_type_override,
                    data_attr=data_attr,
                    field_attr=field_attr,
                    index=index,
                )
                fields.append(field)
            elif isinstance(col_type_override, REVERSE_FOREIGN_KEY_CHOICE_COLUMNS):
                field_name = table_field['field']
                field = self.get_reverse_foreign_key_choice_field(
                    col_type_override=col_type_override,
                    data_attr=data_attr,
                    field_attr=field_attr,
                    index=index,
                )
                fields.append(field)

            elif isinstance(col_type_override, REVERSE_FOREIGN_KEY_DATE_COLUMNS):
                field_name = table_field['field']
                field = self.get_reverse_foreign_key_date_field(
                    col_type_override=col_type_override,
                    data_attr=data_attr,
                    field_attr=field_attr,
                    index=index,
                )
                fields.append(field)

            elif isinstance(col_type_override, LINK_COLUMNS):
                field_name = self.get_link_field(
                    table_field=table_field,
                    col_type_override=col_type_override,
                    fields=fields,
                )

            elif django_field is None and table_field['field'] in [
                'rb_addition',
                'rb_subtraction',
                'rb_times',
                'rb_division',
                'rb_percentage',
            ]:
                self.setup_mathematical_field(
                    data_attr=data_attr,
                    fields=fields,
                    field_attr=field_attr,
                    table_field=table_field,
                    col_type_override=col_type_override,
                    index=index,
                    totals=totals,
                )
            else:
                if col_type_override is not None and data_attr.get('annotation_label') == '1':
                    if isinstance(col_type_override.field, list):
                        table.initial_values += col_type_override.field
                    else:
                        table.initial_values.append(col_type_override.field)

                field_name = field
                if isinstance(col_type_override, CURRENCY_COLUMNS) and totals is not None:
                    annotations_type = int(data_attr.get('annotations_type', 0))
                    append_annotation_query = int(data_attr.get('append_annotation_query', 0)) == 1
                    field_name = self.get_number_fields(
                        field_name=field_name,
                        table=table,
                        base_model=base_model,
                        report_builder_class=report_builder_class,
                        annotations_type=annotations_type,
                        append_annotation_query=append_annotation_query,
                        index=index,
                        data_attr=data_attr,
                        table_field=table_field,
                        fields=fields,
                        totals=totals,
                        col_type_override=col_type_override,
                        decimal_places=2,
                    )

                elif col_type_override.annotations is not None:
                    css_class = col_type_override.column_defs.get('className')
                    show_total = data_attr.get('show_totals')
                    if show_total == '1':
                        self.set_annotation_total(
                            totals=totals,
                            field_name=field_name,
                            col_type_override=col_type_override,
                            decimal_places=2,
                            css_class=css_class,
                        )
                    if field_attr:
                        field = (field, field_attr)
                        fields.append(field)
                else:
                    if field_attr:
                        field = (field, field_attr)
                    fields.append(field)

            if not first_field_name:
                first_field_name = field_name
            fields_map[original_field_name] = field_name

        if not has_annotations and len(report_builder_class.default_columns) > 0:
            table.add_columns(*report_builder_class.default_columns)
        table.add_columns(*fields)
        table.show_pivot_table = False
        if pivot_fields is not None:
            for pivot_field in pivot_fields:
                pivot_field_data = self._get_pivot_details(
                    base_model=base_model,
                    pivot_str=pivot_field['field'],
                    report_builder_class=report_builder_class,
                )
                if pivot_field_data is None:
                    continue

                if pivot_field_data['id'] not in fields_used:
                    table.add_columns('.' + pivot_field_data['id'])
                    fields_used.add(pivot_field_data['id'])

                pivot_field_details = pivot_field_data['details']
                table.add_js_filters(
                    pivot_field_details['type'],
                    pivot_field_data['id'],
                    filter_title=pivot_field_details['title'],
                    **pivot_field_details['kwargs'],
                )
                table.show_pivot_table = True
        if totals:
            totals[first_field_name] = {'text': 'Totals'}
            table.add_plugin(self.column_totals_class, totals)

    def get_number_fields(
        self,
        field_name,
        table,
        base_model,
        report_builder_class,
        annotations_type,
        append_annotation_query,
        index,
        data_attr,
        table_field,
        fields,
        totals,
        col_type_override,
        decimal_places,
    ):
        if (annotations_type != ANNOTATION_CHOICE_NA or append_annotation_query) and data_attr.get(
            'multiple_columns'
        ) == '1':
            query = self.extra_filters(query=table.model.objects)
            multiple_column_field = data_attr.get('multiple_column_field')

            field_report_builder_class = self._get_report_builder_class(
                base_model=base_model,
                field_str=multiple_column_field,
                report_builder_class=report_builder_class,
            )
            _fields = field_report_builder_class.default_multiple_column_fields
            default_multiple_column_fields = [multiple_column_field + '__' + x for x in _fields]
            order_by = f'{multiple_column_field}__{field_report_builder_class.default_multiple_pk}'

            results = (
                query.order_by(order_by)
                .distinct(multiple_column_field)
                .values(multiple_column_field, *default_multiple_column_fields)
            )

            for multiple_index, result in enumerate(results):
                suffix = self._set_multiple_title(
                    database_values=result,
                    value_prefix=multiple_column_field,
                    fields=_fields,
                    text=field_report_builder_class.default_multiple_column_text,
                )
                extra_filter = Q((multiple_column_field, result[multiple_column_field]))

                field_name = self.get_number_field(
                    annotations_type=annotations_type,
                    append_annotation_query=append_annotation_query,
                    index=f'{index}_{multiple_index}',
                    data_attr=data_attr,
                    table_field=table_field,
                    fields=fields,
                    totals=totals,
                    col_type_override=col_type_override,
                    extra_filter=extra_filter,
                    title_suffix=suffix,
                    decimal_places=decimal_places,
                )
        else:
            field_name = self.get_number_field(
                annotations_type=annotations_type,
                append_annotation_query=append_annotation_query,
                index=index,
                data_attr=data_attr,
                table_field=table_field,
                fields=fields,
                totals=totals,
                col_type_override=col_type_override,
                decimal_places=decimal_places,
            )
        return field_name

    @staticmethod
    def _set_multiple_title(database_values, value_prefix, fields, text):
        results = {}
        for field in fields:
            value = database_values[value_prefix + '__' + field]
            results[field] = value
        return text.format(**results)

    def extra_filters(self, query):
        report_query = self.get_report_query(report=self.table_report)
        if report_query:
            query = self.process_query_filters(query=query, search_filter_data=report_query.query)
        return query

    def setup_mathematical_field(
        self,
        data_attr,
        fields,
        field_attr,
        table_field,
        col_type_override,
        index,
        totals,
    ):
        field = table_field['field']
        if field == 'rb_percentage':
            self.get_mathematical_percentage_field(
                fields=fields,
                field_attr=field_attr,
                data_attr=data_attr,
                table_field=table_field,
                index=index,
                totals=totals,
            )
        elif field == 'rb_division':
            values = self.decode_mathematical_columns(
                data_attr=data_attr,
                first_value_column_name='numerator_column',
                second_value_column_name='denominator_column',
            )
            if values is not None:
                expression = ExpressionWrapper(NullIf(F(values[0]), 0) / F(values[1]), output_field=FloatField())
                self.get_mathematical_field(
                    fields=fields,
                    field_attr=field_attr,
                    data_attr=data_attr,
                    table_field=table_field,
                    col_type_override=col_type_override,
                    index=index,
                    totals=totals,
                    expression=expression,
                )
        elif field == 'rb_times':
            values = self.decode_mathematical_columns(
                data_attr=data_attr,
                first_value_column_name='multiplicand_column',
                second_value_column_name='multiplier_column',
            )
            if values is not None:
                expression = ExpressionWrapper(NullIf(F(values[0]), 0) * F(values[1]), output_field=FloatField())
                self.get_mathematical_field(
                    fields=fields,
                    field_attr=field_attr,
                    data_attr=data_attr,
                    table_field=table_field,
                    col_type_override=col_type_override,
                    index=index,
                    totals=totals,
                    expression=expression,
                )
        elif field == 'rb_addition':
            values = self.decode_mathematical_columns(data_attr=data_attr)
            if values is not None:
                expression = ExpressionWrapper(NullIf(F(values[0]), 0) + F(values[1]), output_field=FloatField())
                self.get_mathematical_field(
                    fields=fields,
                    field_attr=field_attr,
                    data_attr=data_attr,
                    table_field=table_field,
                    col_type_override=col_type_override,
                    index=index,
                    totals=totals,
                    expression=expression,
                )
        elif field == 'rb_subtraction':
            values = self.decode_mathematical_columns(data_attr=data_attr)
            if values is not None:
                expression = ExpressionWrapper(NullIf(F(values[0]), 0) - F(values[1]), output_field=FloatField())
                self.get_mathematical_field(
                    fields=fields,
                    field_attr=field_attr,
                    data_attr=data_attr,
                    table_field=table_field,
                    col_type_override=col_type_override,
                    index=index,
                    totals=totals,
                    expression=expression,
                )

    @staticmethod
    def decode_mathematical_columns(
        data_attr,
        first_value_column_name='first_value_column',
        second_value_column_name='second_value_column',
    ):
        first_value_column = data_attr.get(first_value_column_name)
        second_value_column = data_attr.get(second_value_column_name)
        if not first_value_column or not second_value_column:
            return None

        first_value_column = decode_attribute(first_value_column)
        second_value_column = decode_attribute(second_value_column)

        return first_value_column, second_value_column

    def get_mathematical_percentage_field(self, fields, field_attr, data_attr, table_field, index, totals):
        values = self.decode_mathematical_columns(
            data_attr=data_attr,
            first_value_column_name='numerator_column',
            second_value_column_name='denominator_column',
        )
        if values is None:
            return False

        expression = ExpressionWrapper(NullIf(F(values[0]) * 100.00, 0) / F(values[1]), output_field=FloatField())
        decimal_places = data_attr.get('decimal_places', 2)
        field_attr['decimal_places'] = int(decimal_places)
        alignment_class = ALIGNMENT_CLASS.get(int(data_attr.get('alignment', ALIGNMENT_CHOICE_RIGHT)))
        field_attr['options'] = {}
        field = table_field['field']

        column_id = data_attr.get('column_id')
        field_name = decode_attribute(column_id) if column_id else f'{field}_{index}'
        field_attr['annotations'] = {field_name: expression}

        field_attr.update(
            {
                'field': field_name,
                'column_name': field_name,
                'model_path': '',
                'render': [render_replace(html='%1%&thinsp;%', column=field_name)],
                'hidden': data_attr.get('hidden', 0) == '1',
            }
        )
        if alignment_class != '':
            field_attr['column_defs'] = {'className': alignment_class}
        field = self.number_field(**field_attr)

        if totals is not None:
            show_total = data_attr.get('show_totals')
            if show_total == '1':
                self.set_percentage_total(
                    totals=totals,
                    field_name=field_name,
                    denominator=values[1],
                    numerator=values[0],
                    decimal_places=decimal_places,
                    css_class=alignment_class,
                )
        fields.append(field)

    def get_mathematical_field(
        self,
        fields,
        field_attr,
        data_attr,
        table_field,
        col_type_override,
        index,
        totals,
        expression,
    ):
        decimal_places = data_attr.get('decimal_places', 2)
        alignment_class = ALIGNMENT_CLASS.get(int(data_attr.get('alignment', ALIGNMENT_CHOICE_RIGHT)))
        field_attr['decimal_places'] = int(decimal_places)
        field_attr['options'] = {}
        field = table_field['field']
        column_id = data_attr.get('column_id')
        field_name = decode_attribute(column_id) if column_id else f'{field}_{index}'
        field_attr['annotations'] = {field_name: expression}

        field_attr.update(
            {
                'field': field_name,
                'column_name': field_name,
                'model_path': '',
                'trim_zeros': False,
                'hidden': data_attr.get('hidden', 0) == '1',
                'css_class': alignment_class,
            }
        )
        if alignment_class != '':
            field_attr['column_defs'] = {'className': alignment_class}
        field = self.number_field(**field_attr)

        if totals is not None:
            show_total = data_attr.get('show_totals')
            if show_total == '1':
                self.set_number_total(
                    totals=totals,
                    field_name=field_name,
                    col_type_override=col_type_override,
                    decimal_places=decimal_places,
                    css_class=alignment_class,
                )
        fields.append(field)

    def get_reverse_foreign_key_str_field(self, col_type_override, data_attr, field_attr, index):
        col_type_override.table = None
        field_name = f'{col_type_override.field_name}_{index}'
        field = copy.deepcopy(col_type_override)
        delimiter_type = int(data_attr.get('delimiter_type', REVERSE_FOREIGN_KEY_DELIMITER_COMMA))

        sub_query = None
        if int(data_attr.get('has_filter', 0)) == 1:
            _filter = json.loads(decode_attribute(data_attr['filter']))
            prefix_field_name = col_type_override.field_name.split('__')[0]
            sub_query = self.process_filters(search_filter_data=_filter, prefix_field_name=prefix_field_name)
        field.setup_annotations(delimiter_type=delimiter_type, sub_filter=sub_query, field_name=field_name)
        if field_attr:
            field = (field, field_attr)
        return field

    def get_reverse_foreign_key_bool_field(self, col_type_override, data_attr, field_attr, index):
        col_type_override.table = None
        field_name = f'{col_type_override.field_name}_{index}'
        field = copy.deepcopy(col_type_override)
        sub_query = None
        if int(data_attr.get('has_filter', 0)) == 1:
            _filter = json.loads(decode_attribute(data_attr['filter']))
            prefix_field_name = col_type_override.field_name.split('__')[0]
            sub_query = self.process_filters(search_filter_data=_filter, prefix_field_name=prefix_field_name)
        annotations_type = int(data_attr.get('annotations_type', REVERSE_FOREIGN_KEY_ANNOTATION_BOOLEAN_XOR))
        field.setup_annotations(
            annotations_type=annotations_type,
            sub_filter=sub_query,
            field_name=field_name,
        )
        if field_attr:
            field = (field, field_attr)
        return field

    def get_reverse_foreign_key_choice_field(self, col_type_override, data_attr, field_attr, index):
        col_type_override.table = None
        field_name = f'{col_type_override.field_name}_{index}'
        field = copy.deepcopy(col_type_override)
        sub_query = None
        delimiter_type = int(data_attr.get('delimiter_type', REVERSE_FOREIGN_KEY_DELIMITER_COMMA))
        if int(data_attr.get('has_filter', 0)) == 1:
            _filter = json.loads(decode_attribute(data_attr['filter']))
            prefix_field_name = col_type_override.field_name.split('__')[0]
            sub_query = self.process_filters(search_filter_data=_filter, prefix_field_name=prefix_field_name)
        field.setup_annotations(delimiter_type=delimiter_type, sub_filter=sub_query, field_name=field_name)
        if field_attr:
            field = (field, field_attr)
        return field

    def get_reverse_foreign_key_date_field(self, col_type_override, data_attr, field_attr, index):
        col_type_override.table = None
        field_name = f'{col_type_override.field_name}_{index}'
        field = copy.deepcopy(col_type_override)
        sub_query = None
        delimiter_type = int(data_attr.get('delimiter_type', REVERSE_FOREIGN_KEY_DELIMITER_COMMA))
        date_format_type = int(data_attr.get('date_format', DATE_FORMAT_TYPE_DD_MM_YY_SLASH))
        annotations_type = int(data_attr.get('annotations_type', REVERSE_FOREIGN_KEY_ANNOTATION_BOOLEAN_XOR))

        if int(data_attr.get('has_filter', 0)) == 1:
            _filter = json.loads(decode_attribute(data_attr['filter']))
            prefix_field_name = col_type_override.field_name.split('__')[0]
            sub_query = self.process_filters(search_filter_data=_filter, prefix_field_name=prefix_field_name)

        field.setup_annotations(
            delimiter_type=delimiter_type,
            date_format_type=date_format_type,
            annotations_type=annotations_type,
            sub_filter=sub_query,
            field_name=field_name,
        )
        if field_attr:
            field = (field, field_attr)
        return field


class DescriptionColumn(ColumnBase):
    def row_result(self, data, _page_data, columns):
        html = self.options['html']

        for column in columns:
            if not isinstance(column, DescriptionColumn):
                data[column.column_name] = column.row_result(data, _page_data)

        try:
            template = Template(html)
            context = Context(data)
            return template.render(context)
        except TemplateSyntaxError as e:
            return f'Error in description ({e})'



from ajax_helpers.mixins import AjaxHelpers
from django_menus.menu import MenuItem, MenuMixin

from advanced_report_builder.utils import make_slug_str


class ReportBase(AjaxHelpers, MenuMixin):
    # noinspection PyUnusedLocal
    @staticmethod
    def duplicate_menu(request, report_id):
        return [
            MenuItem(
                f'advanced_report_builder:duplicate_report_modal,pk-{report_id}',
                css_classes=['btn-success'],
            )
        ]

    def get_dashboard_class(self, report):
        return None

    def queries_menu(self, report, dashboard_report):
        query_slug = f'query{report.id}'
        if dashboard_report is not None:
            if not dashboard_report.show_versions:
                return []
            query_slug += f'_{dashboard_report.id}'

        report_queries = report.reportquery_set.all()
        if len(report_queries) > 1:
            dropdown = []
            for report_query in report_queries:
                slug_str = make_slug_str(self.slug, overrides={query_slug: report_query.id})
                dropdown.append(
                    (
                        self.request.resolver_match.view_name,
                        report_query.name,
                        {'url_kwargs': {'slug': slug_str}},
                    )
                )
            return [
                MenuItem(
                    menu_display='Version',
                    no_hover=True,
                    css_classes='btn-secondary',
                    dropdown=dropdown,
                )
            ]
        return []



# Generated by Django 5.0.7 on 2025-05-07 11:13

import django.db.models.deletion
from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('advanced_report_builder', '0014_auto_20240706_2207'),
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.AddField(
            model_name='report',
            name='user_created',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='report_user_created_set', to=settings.AUTH_USER_MODEL),
        ),
        migrations.AddField(
            model_name='report',
            name='user_updated',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='report_user_updated_set', to=settings.AUTH_USER_MODEL),
        ),
    ]



from django.db.models import Func
from django.db.models.expressions import RawSQL


class GenerateSeries(Func):
    function = 'GENERATE_SERIES'

    def __init__(self, start_date_field, end_date_field, interval='1 month', output_field=None):
        expressions = [
            start_date_field,
            end_date_field,
            RawSQL('%s', [interval]),
        ]
        super().__init__(*expressions, output_field=output_field)



import base64
import json
import operator
from datetime import datetime, timedelta
from functools import reduce

from crispy_forms.bootstrap import StrictButton
from crispy_forms.layout import Div
from date_offset.monthdelta import MonthDelta
from django.core.exceptions import FieldDoesNotExist
from django.db.models import Q
from django.forms import BooleanField, CharField, ChoiceField
from django.shortcuts import get_object_or_404
from django.urls import reverse
from django_datatables.datatables import DatatableTable
from django_menus.menu import MenuItem
from django_modals.fields import FieldEx
from django_modals.form_helpers import HorizontalNoEnterHelper
from django_modals.modals import FormModal, Modal
from django_modals.processes import PERMISSION_OFF, PROCESS_EDIT_DELETE
from django_modals.widgets.colour_picker import ColourPickerWidget
from django_modals.widgets.select2 import Select2, Select2Multiple
from django_modals.widgets.widgets import Toggle

from advanced_report_builder.column_types import NUMBER_FIELDS
from advanced_report_builder.exceptions import ReportError
from advanced_report_builder.generate_series import GenerateSeries
from advanced_report_builder.globals import (
    ANNOTATION_VALUE_DAY,
    ANNOTATION_VALUE_FUNCTIONS,
    ANNOTATION_VALUE_MONTH,
    ANNOTATION_VALUE_QUARTER,
    ANNOTATION_VALUE_WEEK,
    ANNOTATION_VALUE_YEAR,
    DATE_FORMAT_TYPES_DJANGO_FORMAT,
    DEFAULT_DATE_FORMAT,
    GENERATE_SERIES_INTERVALS,
)
from advanced_report_builder.models import BarChartReport, ReportType
from advanced_report_builder.toggle import RBToggle
from advanced_report_builder.utils import (
    decode_attribute,
    encode_attribute,
    get_report_builder_class,
    split_attr,
)
from advanced_report_builder.views.charts_base import ChartBaseFieldForm, ChartBaseView
from advanced_report_builder.views.datatables.modal import (
    TableFieldForm,
    TableFieldModal,
)
from advanced_report_builder.views.datatables.utils import TableUtilsMixin
from advanced_report_builder.views.modals_base import (
    QueryBuilderModalBase,
    QueryBuilderModalBaseMixin,
)
from advanced_report_builder.views.query_modal.mixin import MultiQueryModalMixin


class BarChartView(ChartBaseView):
    def dispatch(self, request, *args, **kwargs):
        self.report = kwargs.get('report')
        self.chart_report = self.report.barchartreport
        return super().dispatch(request, *args, **kwargs)

    def get_context_data(self, **kwargs):
        try:
            context = super().get_context_data(**kwargs)
        except FieldDoesNotExist as e:
            raise ReportError(e)
        self.table.bar_chart_report = self.chart_report
        self.table.datatable_template = 'advanced_report_builder/charts/bar/middle.html'
        self.table.breakdown_url = self.get_breakdown_url()

        context['bar_chart_report'] = self.chart_report
        return context

    def get_breakdown_url(self):
        if self.table.bar_chart_report.show_breakdown:
            enable_links = self.slug.get('enable_links') == 'True'
            slug = f'pk-{self.table.bar_chart_report.id}-data-99999-date-88888-enable_links-{enable_links}'
            return reverse(
                'advanced_report_builder:bar_chart_show_breakdown_modal',
                kwargs={'slug': slug},
            )
        return None

    def get_date_format(self):
        if self.chart_report.show_blank_dates:
            date_format = '%Y-%m-%d'
        else:
            default_format_type = DEFAULT_DATE_FORMAT[self.chart_report.axis_scale]
            date_format = DATE_FORMAT_TYPES_DJANGO_FORMAT[default_format_type]
        return date_format

    def set_extra_number_field_kwargs(self, data_attr, options, multiple_index, additional_options):
        negative_bar_colour = None
        positive_bar_colour = None
        if additional_options is not None:
            negative_bar_colour = additional_options.get('negative_colour')
            positive_bar_colour = additional_options.get('positive_colour')
        if negative_bar_colour is None or negative_bar_colour == '':
            negative_bar_colour = data_attr.get('negative_bar_colour') or '801C70'
            negative_bar_colour = self.add_colour_offset(negative_bar_colour, multiple_index=multiple_index)
        if positive_bar_colour is None or positive_bar_colour == '':
            positive_bar_colour = data_attr.get('positive_bar_colour') or '801C70'
            positive_bar_colour = self.add_colour_offset(positive_bar_colour, multiple_index=multiple_index)

        options.update(
            {
                'colours': {
                    'negative': negative_bar_colour,
                    'positive': positive_bar_colour,
                }
            }
        )

    def edit_report_menu(self, request, chart_report_id, slug_str):
        return [
            MenuItem(
                f'advanced_report_builder:bar_chart_modal,pk-{chart_report_id}{slug_str}',
                menu_display='Edit',
                font_awesome='fas fa-pencil-alt',
                css_classes=['btn-primary'],
            ),
            *self.duplicate_menu(request=self.request, report_id=chart_report_id),
        ]

    def get_date_field(self, index, fields, base_model, table):
        if self.chart_report.date_field_type == BarChartReport.DATE_FIELD_SINGLE:
            return super().get_date_field(index=index, fields=fields, base_model=base_model, table=table)

        start_field_name = self.chart_report.date_field
        end_field_name = self.chart_report.end_date_field

        if start_field_name is None or end_field_name is None:
            return

        report_builder_class = get_report_builder_class(model=base_model, report_type=self.chart_report.report_type)

        start_django_field, start_col_type_override, _, _ = self.get_field_details(
            base_model=base_model,
            field=start_field_name,
            report_builder_class=report_builder_class,
            table=table,
        )
        if start_col_type_override:
            start_field_name = start_col_type_override.field

        end_django_field, end_col_type_override, _, _ = self.get_field_details(
            base_model=base_model,
            field=end_field_name,
            report_builder_class=report_builder_class,
            table=table,
        )
        if end_col_type_override:
            end_field_name = end_col_type_override.field

        date_format = self.get_date_format()

        date_function_kwargs = {'title': start_field_name, 'date_format': date_format}

        annotations_value = self.chart_report.axis_scale

        new_field_name = f'{annotations_value}_{start_field_name}_{index}'
        function = ANNOTATION_VALUE_FUNCTIONS[annotations_value]
        date_function_kwargs['annotations_value'] = {
            new_field_name: GenerateSeries(
                start_date_field=function(start_field_name),
                end_date_field=function(end_field_name),
                interval=GENERATE_SERIES_INTERVALS[annotations_value],
            )
        }
        start_field_name = new_field_name

        date_function_kwargs.update(
            {
                'field': start_field_name,
                'column_name': start_field_name,
                'model_path': '',
            }
        )

        field = self.date_field(**date_function_kwargs)
        fields.append(field)
        return start_field_name


class BarChartModal(MultiQueryModalMixin, QueryBuilderModalBase):
    template_name = 'advanced_report_builder/charts/modal.html'
    process = PROCESS_EDIT_DELETE
    permission_delete = PERMISSION_OFF
    model = BarChartReport
    show_order_by = False
    widgets = {
        'positive_bar_colour': ColourPickerWidget,
        'negative_bar_colour': ColourPickerWidget,
        'stacked': RBToggle,
        'show_totals': RBToggle,
        'show_blank_dates': RBToggle,
        'date_field_type': Select2,
        'report_tags': Select2Multiple,
        'show_breakdown': Toggle(attrs={'data-onstyle': 'success', 'data-on': 'YES', 'data-off': 'NO'}),
    }

    form_fields = [
        'name',
        'notes',
        'report_type',
        'report_tags',
        ('bar_chart_orientation', {'label': 'Orientation'}),
        'axis_value_type',
        'axis_scale',
        'date_field_type',
        'date_field',
        'end_date_field',
        'fields',
        'x_label',
        'y_label',
        'stacked',
        'show_totals',
        'show_blank_dates',
        'show_breakdown',
        'breakdown_fields',
    ]

    def form_setup(self, form, *_args, **_kwargs):
        if 'data' in _kwargs and len(_kwargs['data']) > 0:
            date_field = _kwargs['data'].get('date_field')
            end_date_field = _kwargs['data'].get('end_date_field')
            report_type_id = _kwargs['data'].get('report_type')
            report_type = get_object_or_404(ReportType, id=report_type_id)
        else:
            date_field = form.instance.date_field
            end_date_field = form.instance.end_date_field
            report_type = form.instance.report_type

        self.setup_field(
            field_type='date',
            form=form,
            field_name='date_field',
            selected_field_id=date_field,
            report_type=report_type,
        )

        self.setup_field(
            field_type='date',
            form=form,
            field_name='end_date_field',
            selected_field_id=end_date_field,
            report_type=report_type,
        )

        form.fields['notes'].widget.attrs['rows'] = 3

        url = reverse(
            'advanced_report_builder:bar_chart_field_modal',
            kwargs={'slug': 'selector-99999-data-FIELD_INFO-report_type_id-REPORT_TYPE_ID'},
        )

        url_breakdown = reverse(
            'advanced_report_builder:bar_chart_breakdown_field_modal',
            kwargs={'slug': 'selector-99999-data-FIELD_INFO-report_type_id-REPORT_TYPE_ID'},
        )

        form.add_trigger(
            'show_breakdown',
            'onchange',
            [
                {
                    'selector': '#div_id_breakdown_fields',
                    'values': {'checked': 'show'},
                    'default': 'hide',
                },
            ],
        )

        form.add_trigger(
            'date_field_type',
            'onchange',
            [
                {
                    'selector': '#div_id_end_date_field',
                    'values': {BarChartReport.DATE_FIELD_RANGE: 'show'},
                    'default': 'hide',
                },
                {
                    'selector': 'label[for=id_date_field]',
                    'values': {
                        BarChartReport.DATE_FIELD_SINGLE: ('html', 'Date field'),
                        BarChartReport.DATE_FIELD_RANGE: ('html', 'Start date field'),
                    },
                },
            ],
        )

        fields = [
            'name',
            'notes',
            'report_type',
            'report_tags',
            'bar_chart_orientation',
            'axis_scale',
            'axis_value_type',
            'date_field_type',
            'date_field',
            'end_date_field',
            FieldEx(
                'fields',
                template='advanced_report_builder/select_column.html',
                extra_context={'select_column_url': url, 'command_prefix': ''},
            ),
            'x_label',
            'y_label',
            'stacked',
            'show_totals',
            'show_blank_dates',
            'show_breakdown',
            FieldEx(
                'breakdown_fields',
                template='advanced_report_builder/select_column.html',
                extra_context={
                    'select_column_url': url_breakdown,
                    'command_prefix': 'breakdown_',
                },
            ),
        ]
        if self.object.id:
            self.add_extra_queries(form=form, fields=fields)
        return fields

    def select2_date_field(self, **kwargs):
        return self.get_fields_for_select2(
            field_type='date',
            report_type=kwargs['report_type'],
            search_string=kwargs.get('search'),
        )

    def select2_end_date_field(self, **kwargs):
        return self.select2_date_field(**kwargs)

    def _get_ajax_fields(self, report_type, prefix=''):
        report_type_id = report_type
        report_builder_fields, base_model = self.get_report_builder_class(report_type_id=report_type_id)
        fields = []
        tables = []
        self._get_fields(
            base_model=base_model,
            fields=fields,
            tables=tables,
            report_builder_class=report_builder_fields,
        )

        self.add_command(
            f'{prefix}report_fields',
            data=json.dumps({'fields': fields, 'tables': tables}),
        )
        return self.command_response()

    def ajax_get_fields(self, **kwargs):
        return self._get_ajax_fields(report_type=kwargs['report_type'])

    def ajax_get_breakdown_fields(self, **kwargs):
        return self._get_ajax_fields(report_type=kwargs['report_type'], prefix='breakdown_')

    def clean(self, form, cleaned_data):
        end_date_field = cleaned_data['end_date_field']
        if cleaned_data['date_field_type'] == BarChartReport.DATE_FIELD_RANGE and (
            end_date_field is None or end_date_field == ''
        ):
            form.add_error('end_date_field', 'Please select a date')


class BarChartFieldForm(ChartBaseFieldForm):
    def setup_modal(self, *args, **kwargs):
        data = json.loads(base64.b64decode(self.slug['data']))
        report_type, base_model = self.get_report_type_details()

        data_attr = split_attr(data)

        self.fields['title'] = CharField(initial=data['title'])

        self.fields['positive_bar_colour'] = CharField(required=False, widget=ColourPickerWidget)
        self.fields['negative_bar_colour'] = CharField(required=False, widget=ColourPickerWidget)
        self.fields['positive_bar_colour'].initial = data_attr.get('positive_bar_colour')
        self.fields['negative_bar_colour'].initial = data_attr.get('negative_bar_colour')

        self.fields['has_filter'] = BooleanField(required=False, widget=RBToggle())
        self.fields['filter'] = CharField(required=False)

        if data_attr.get('has_filter') == '1':
            self.fields['has_filter'].initial = True
            if 'filter' in data_attr:
                self.fields['filter'].initial = decode_attribute(data_attr['filter'])

        self.fields['multiple_columns'] = BooleanField(required=False, widget=RBToggle())
        self.fields['append_column_title'] = BooleanField(required=False, widget=RBToggle())
        report_builder_class = get_report_builder_class(model=base_model, report_type=report_type)

        self.setup_colour_field(
            form_fields=self.fields,
            base_model=base_model,
            report_builder_class=report_builder_class,
            name='positive_colour_field',
            data_attr=data_attr,
        )
        self.setup_colour_field(
            form_fields=self.fields,
            base_model=base_model,
            report_builder_class=report_builder_class,
            name='negative_colour_field',
            data_attr=data_attr,
        )

        multiple_column_field = []
        self._get_query_builder_foreign_key_fields(
            base_model=base_model,
            report_builder_class=report_builder_class,
            fields=multiple_column_field,
        )
        self.fields['multiple_column_field'] = ChoiceField(
            choices=multiple_column_field, required=False, widget=Select2()
        )

        if data_attr.get('multiple_columns') == '1':
            self.fields['multiple_columns'].initial = True
            self.fields['multiple_column_field'].initial = data_attr.get('multiple_column_field')

            if data_attr.get('append_column_title') == '1':
                self.fields['append_column_title'].initial = True

        super().setup_modal(*args, **kwargs)

    def get_additional_attributes(self):
        attributes = []
        self.get_report_type_details()

        attributes.append(f'positive_bar_colour-{self.cleaned_data["positive_bar_colour"]}')
        attributes.append(f'negative_bar_colour-{self.cleaned_data["negative_bar_colour"]}')

        if self.cleaned_data['has_filter']:
            attributes.append('has_filter-1')

            if self.cleaned_data['filter']:
                b64_filter = encode_attribute(self.cleaned_data['filter'])
                attributes.append(f'filter-{b64_filter}')

            if self.cleaned_data['multiple_columns']:
                attributes.append('multiple_columns-1')

                if self.cleaned_data['multiple_column_field']:
                    attributes.append(f'multiple_column_field-{self.cleaned_data["multiple_column_field"]}')

                if self.cleaned_data['append_column_title']:
                    attributes.append('append_column_title-1')

                if self.cleaned_data['positive_colour_field']:
                    attributes.append(f'positive_colour_field-{self.cleaned_data["positive_colour_field"]}')

                if self.cleaned_data['negative_colour_field']:
                    attributes.append(f'negative_colour_field-{self.cleaned_data["negative_colour_field"]}')

        if attributes:
            return '-'.join(attributes)
        return None


class BarChartFieldModal(QueryBuilderModalBaseMixin, FormModal):
    form_class = BarChartFieldForm
    size = 'xl'
    template_name = 'advanced_report_builder/charts/modal_field.html'
    no_header_x = True
    helper_class = HorizontalNoEnterHelper

    @property
    def modal_title(self):
        data = json.loads(base64.b64decode(self.slug['data']))
        return f'Edit {data["title"]}'

    def form_valid(self, form):
        selector = self.slug['selector']

        _attr = form.get_additional_attributes()
        self.add_command(
            {
                'function': 'set_attr',
                'selector': f'#{selector}',
                'attr': 'data-attr',
                'val': _attr,
            }
        )

        self.add_command(
            {
                'function': 'html',
                'selector': f'#{selector} span',
                'html': form.cleaned_data['title'],
            }
        )
        self.add_command({'function': 'update_selection'})
        return self.command_response('close')

    # noinspection PyMethodMayBeStatic
    def form_setup(self, form, *_args, **_kwargs):
        form.add_trigger(
            'has_filter',
            'onchange',
            [
                {
                    'selector': '#filter_fields_div',
                    'values': {'checked': 'show'},
                    'default': 'hide',
                }
            ],
        )

        form.add_trigger(
            'multiple_columns',
            'onchange',
            [
                {
                    'selector': '#multiple_columns_fields_div',
                    'values': {'checked': 'show'},
                    'default': 'hide',
                },
            ],
        )

        return [
            'title',
            'positive_bar_colour',
            'negative_bar_colour',
            Div(
                FieldEx(
                    'has_filter',
                    template='django_modals/fields/label_checkbox.html',
                    field_class='col-6 input-group-sm',
                ),
                Div(
                    FieldEx(
                        'filter',
                        template='advanced_report_builder/datatables/fields/single_query_builder.html',
                    ),
                    FieldEx(
                        'multiple_columns',
                        template='django_modals/fields/label_checkbox.html',
                        field_class='col-6 input-group-sm',
                    ),
                    Div(
                        FieldEx(
                            'append_column_title',
                            template='django_modals/fields/label_checkbox.html',
                            field_class='col-6 input-group-sm',
                        ),
                        FieldEx('multiple_column_field'),
                        FieldEx('positive_colour_field'),
                        FieldEx('negative_colour_field'),
                        css_id='multiple_columns_fields_div',
                    ),
                    css_id='filter_fields_div',
                ),
                css_id='annotations_fields_div',
            ),
        ]

    def ajax_get_query_builder_fields(self, **kwargs):
        field_auto_id = kwargs['field_auto_id']

        report_type_id = self.slug['report_type_id']
        query_builder_filters = self.get_query_builder_report_type_field(report_type_id=report_type_id)

        return self.command_response(f'query_builder_{field_auto_id}', data=json.dumps(query_builder_filters))


class BarChartBreakdownFieldForm(TableFieldForm):
    cancel_class = 'btn-secondary modal-cancel'

    def submit_button(self, css_class='btn-success modal-submit', button_text='Submit', **kwargs):
        if isinstance(self.django_field, NUMBER_FIELDS):
            return StrictButton(
                button_text,
                onclick=f'save_modal_{self.form_id}()',
                css_class=css_class,
                **kwargs,
            )
        else:
            return super().submit_button(css_class, button_text, **kwargs)


class BarChartBreakdownFieldModal(TableFieldModal):
    form_class = BarChartBreakdownFieldForm
    update_selection_command = 'breakdown_update_selection'


class BarChartShowBreakdownModal(TableUtilsMixin, Modal):
    button_container_class = 'text-center'
    size = 'xl'

    def __init__(self, *args, **kwargs):
        self.date_field_path = None
        self.end_date_field_path = None
        self.bar_chart_report = None
        self.chart_report = None
        self.field_filter = None
        super().__init__(*args, **kwargs)

    def get_bar_chart_report(self):
        if self.bar_chart_report is None:
            self.bar_chart_report = get_object_or_404(BarChartReport, pk=self.slug['pk'])
        return self.bar_chart_report

    def modal_title(self):
        bar_chart_report = self.get_bar_chart_report()

        data_index = int(self.slug['data'])
        field = bar_chart_report.fields[data_index]
        title = field['title']
        date_title = self.get_date_title()
        return f'{bar_chart_report.name} - {title} - {date_title}'

    def add_table(self, base_model):
        return DatatableTable(view=self, model=base_model)

    def setup_table(self):
        bar_chart_report = self.get_bar_chart_report()
        self.chart_report = bar_chart_report
        self.kwargs['enable_links'] = self.slug['enable_links'] == 'True'
        base_model = bar_chart_report.get_base_model()
        chart_fields = bar_chart_report.fields
        report_builder_class = get_report_builder_class(model=base_model, report_type=bar_chart_report.report_type)
        table = self.add_table(base_model=base_model)

        table_fields = bar_chart_report.breakdown_fields
        fields_used = set()
        fields_map = {}
        self.process_query_results(
            report_builder_class=report_builder_class,
            table=table,
            base_model=base_model,
            fields_used=fields_used,
            fields_map=fields_map,
            table_fields=table_fields,
        )
        data_index = int(self.slug['data'])
        table_field = chart_fields[data_index]
        field = table_field['field']
        django_field, col_type_override, _, _ = self.get_field_details(
            base_model=base_model,
            field=field,
            report_builder_class=report_builder_class,
            table=table,
        )

        self.field_filter = None
        if isinstance(django_field, NUMBER_FIELDS) or (col_type_override is not None and col_type_override.annotations):
            data_attr = split_attr(table_field)

            annotations_type = self.chart_report.axis_value_type
            if annotations_type != 0:
                b64_filter = data_attr.get('filter')
                if b64_filter:
                    _filter = decode_attribute(b64_filter)
                    self.field_filter = json.loads(_filter)

        self.date_field_path = self.get_field_details(
            base_model=base_model,
            field=bar_chart_report.date_field,
            report_builder_class=report_builder_class,
            table=table,
        )[3]
        if bar_chart_report.date_field_type == BarChartReport.DATE_FIELD_RANGE:
            self.end_date_field_path = self.get_field_details(
                base_model=base_model,
                field=bar_chart_report.end_date_field,
                report_builder_class=report_builder_class,
                table=table,
            )[3]

        table.extra_filters = self.extra_filters

        table.ajax_data = False
        table.table_options['pageLength'] = 25
        table.table_options['bStateSave'] = False
        return table

    def modal_content(self):
        table = self.setup_table()
        return table.render()

    def get_date_title(self):
        bar_chart_report = self.bar_chart_report
        start_date = datetime.strptime(self.slug['date'], '%Y_%m_%d').date()

        if bar_chart_report.axis_scale == ANNOTATION_VALUE_YEAR:
            return start_date.year
        elif bar_chart_report.axis_scale in [
            ANNOTATION_VALUE_QUARTER,
            ANNOTATION_VALUE_MONTH,
        ]:
            return start_date.strftime('%b %Y')
        elif bar_chart_report.axis_scale == ANNOTATION_VALUE_WEEK:
            end_date = start_date + timedelta(weeks=1)
            return start_date.strftime('%d/%m/%Y') + ' - ' + end_date.strftime('%d/%b/%Y')

        elif bar_chart_report.axis_scale == ANNOTATION_VALUE_DAY:
            return start_date.strftime('%d/%m/%Y')
        else:
            raise AssertionError()

    def filter_date(self, query):
        bar_chart_report = self.chart_report
        start_date = datetime.strptime(self.slug['date'], '%Y_%m_%d').date()

        if bar_chart_report.axis_scale == ANNOTATION_VALUE_YEAR:
            end_date = start_date + MonthDelta(months=12)
        elif bar_chart_report.axis_scale == ANNOTATION_VALUE_QUARTER:
            end_date = start_date + MonthDelta(months=3)
        elif bar_chart_report.axis_scale == ANNOTATION_VALUE_MONTH:
            end_date = start_date + MonthDelta(months=1)
        elif bar_chart_report.axis_scale == ANNOTATION_VALUE_WEEK:
            end_date = start_date + timedelta(weeks=1)
        elif bar_chart_report.axis_scale == ANNOTATION_VALUE_DAY:
            end_date = start_date + timedelta(days=1)
        else:
            raise AssertionError()

        if self.end_date_field_path is None:
            query_list = [
                (Q((self.date_field_path + '__gte', start_date))) & (Q((self.date_field_path + '__lt', end_date)))
            ]
        else:
            query_list = [
                (Q((self.date_field_path + '__lt', end_date))) & (Q((self.end_date_field_path + '__gte', start_date)))
            ]

        return query.filter(reduce(operator.and_, query_list))

    def extra_filters(self, query):
        report_query = self.get_report_query(report=self.chart_report)
        if report_query:
            query = self.process_query_filters(query=query, search_filter_data=report_query.query)
        if self.field_filter is not None:
            query = self.process_query_filters(query=query, search_filter_data=self.field_filter)

        query = self.filter_date(query=query)
        return query



# Generated by Django 5.0.7 on 2025-05-29 22:03

import django.db.models.deletion
import time_stamped_model.models
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('advanced_report_builder', '0015_report_user_created_report_user_updated'),
    ]

    operations = [
        migrations.CreateModel(
            name='CalendarReport',
            fields=[
                ('report_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, serialize=False, to='advanced_report_builder.report')),
                ('height', models.PositiveSmallIntegerField(default=600)),
                ('view_type', models.PositiveSmallIntegerField(choices=[(1, 'Month'), (2, 'Grid Week'), (3, 'List Week'), (4, 'Day'), (5, 'Year')], default=1)),
            ],
            options={
                'abstract': False,
            },
            bases=('advanced_report_builder.report',),
        ),
        migrations.AddField(
            model_name='dashboardreport',
            name='options',
            field=models.JSONField(blank=True, null=True),
        ),
        migrations.CreateModel(
            name='CalendarReportDescription',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', time_stamped_model.models.CreationDateTimeField(auto_now_add=True)),
                ('modified', time_stamped_model.models.ModificationDateTimeField(auto_now=True)),
                ('name', models.CharField(max_length=200)),
                ('description', models.TextField(blank=True, null=True)),
                ('order', models.PositiveSmallIntegerField()),
                ('calendar_report', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='advanced_report_builder.calendarreport')),
                ('report_type', models.ForeignKey(null=True, on_delete=django.db.models.deletion.PROTECT, to='advanced_report_builder.reporttype')),
            ],
            options={
                'ordering': ('order',),
            },
        ),
        migrations.CreateModel(
            name='CalendarReportDataSet',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', time_stamped_model.models.CreationDateTimeField(auto_now_add=True)),
                ('modified', time_stamped_model.models.ModificationDateTimeField(auto_now=True)),
                ('order', models.PositiveSmallIntegerField()),
                ('heading_field', models.CharField(blank=True, max_length=200, null=True)),
                ('name', models.CharField(max_length=200)),
                ('display_type', models.PositiveSmallIntegerField(choices=[(1, 'Name and Description'), (2, 'Description Only'), (3, 'Heading Only')], default=1)),
                ('start_date_field', models.CharField(blank=True, max_length=200, null=True)),
                ('end_date_type', models.PositiveSmallIntegerField(choices=[(1, 'Field'), (2, 'Duration Field'), (3, 'Duration Fixed')], default=1)),
                ('end_date_field', models.CharField(blank=True, max_length=200, null=True)),
                ('end_duration_field', models.CharField(blank=True, max_length=200, null=True)),
                ('end_duration', models.PositiveSmallIntegerField(blank=True, null=True)),
                ('background_colour_field', models.CharField(blank=True, max_length=200, null=True)),
                ('link_field', models.CharField(blank=True, max_length=200, null=True)),
                ('query_data', models.JSONField(blank=True, null=True)),
                ('calendar_report', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='advanced_report_builder.calendarreport')),
                ('report_type', models.ForeignKey(null=True, on_delete=django.db.models.deletion.PROTECT, to='advanced_report_builder.reporttype')),
                ('calendar_report_description', models.ForeignKey(null=True, on_delete=django.db.models.deletion.CASCADE, to='advanced_report_builder.calendarreportdescription')),
            ],
            options={
                'ordering': ('order',),
            },
        ),
    ]



import base64
import json

from crispy_forms.layout import Div
from django.forms import BooleanField, CharField, ChoiceField
from django.urls import reverse
from django_menus.menu import MenuItem
from django_modals.fields import FieldEx
from django_modals.form_helpers import HorizontalNoEnterHelper
from django_modals.modals import FormModal
from django_modals.processes import PERMISSION_OFF, PROCESS_EDIT_DELETE
from django_modals.widgets.colour_picker import ColourPickerWidget
from django_modals.widgets.select2 import Select2Multiple

from advanced_report_builder.models import PieChartReport
from advanced_report_builder.toggle import RBToggle
from advanced_report_builder.utils import (
    decode_attribute,
    encode_attribute,
    get_report_builder_class,
    split_attr,
)
from advanced_report_builder.views.charts_base import ChartBaseFieldForm, ChartBaseView
from advanced_report_builder.views.modals_base import (
    QueryBuilderModalBase,
    QueryBuilderModalBaseMixin,
)
from advanced_report_builder.views.query_modal.mixin import MultiQueryModalMixin


class PieChartView(ChartBaseView):
    use_annotations = False

    def dispatch(self, request, *args, **kwargs):
        self.report = kwargs.get('report')
        self.chart_report = self.report.piechartreport
        return super().dispatch(request, *args, **kwargs)

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        self.table.pie_chart_report = self.chart_report
        self.table.datatable_template = 'advanced_report_builder/charts/pie/middle.html'
        context['pie_chart_report'] = self.chart_report
        return context

    def set_extra_number_field_kwargs(self, data_attr, options, multiple_index, additional_options):
        pie_colour = None
        if additional_options is not None:
            pie_colour = additional_options.get('positive_colour')
        if pie_colour is None or pie_colour == '':
            pie_colour = data_attr.get('pie_colour') or '801C70'
            pie_colour = self.add_colour_offset(pie_colour, multiple_index=multiple_index)
        options.update({'colour': pie_colour})

    def edit_report_menu(self, request, chart_report_id, slug_str):
        return [
            MenuItem(
                f'advanced_report_builder:pie_chart_modal,pk-{chart_report_id}{slug_str}',
                menu_display='Edit',
                font_awesome='fas fa-pencil-alt',
                css_classes=['btn-primary'],
            ),
            *self.duplicate_menu(request=self.request, report_id=chart_report_id),
        ]

    def get_date_field(self, index, fields, base_model, table):
        return None


class PieChartModal(MultiQueryModalMixin, QueryBuilderModalBase):
    template_name = 'advanced_report_builder/charts/modal.html'
    process = PROCESS_EDIT_DELETE
    permission_delete = PERMISSION_OFF
    model = PieChartReport
    show_order_by = False

    form_fields = [
        'name',
        'notes',
        'report_type',
        'report_tags',
        'axis_value_type',
        'style',
        'fields',
    ]

    widgets = {'report_tags': Select2Multiple}

    def form_setup(self, form, *_args, **_kwargs):
        url = reverse(
            'advanced_report_builder:pie_chart_field_modal',
            kwargs={'slug': 'selector-99999-data-FIELD_INFO-report_type_id-REPORT_TYPE_ID'},
        )
        form.fields['notes'].widget.attrs['rows'] = 3
        fields = [
            'name',
            'notes',
            'report_type',
            'report_tags',
            'axis_value_type',
            'style',
            FieldEx(
                'fields',
                template='advanced_report_builder/select_column.html',
                extra_context={'select_column_url': url, 'command_prefix': ''},
            ),
        ]
        if self.object.id:
            self.add_extra_queries(form=form, fields=fields)
        return fields


class PieChartFieldForm(ChartBaseFieldForm):
    def setup_modal(self, *args, **kwargs):
        data = json.loads(base64.b64decode(self.slug['data']))
        report_type, base_model = self.get_report_type_details()

        data_attr = split_attr(data)

        self.fields['title'] = CharField(initial=data['title'])

        self.fields['pie_colour'] = CharField(required=False, widget=ColourPickerWidget)
        self.fields['pie_colour'].initial = data_attr.get('pie_colour')

        self.fields['has_filter'] = BooleanField(required=False, widget=RBToggle())
        self.fields['filter'] = CharField(required=False)

        if data_attr.get('has_filter') == '1':
            self.fields['has_filter'].initial = True
            if 'filter' in data_attr:
                self.fields['filter'].initial = decode_attribute(data_attr['filter'])

        self.fields['multiple_columns'] = BooleanField(required=False, widget=RBToggle())
        self.fields['append_column_title'] = BooleanField(required=False, widget=RBToggle())
        report_builder_class = get_report_builder_class(model=base_model, report_type=report_type)

        self.setup_colour_field(
            form_fields=self.fields,
            base_model=base_model,
            report_builder_class=report_builder_class,
            name='positive_colour_field',
            data_attr=data_attr,
            label='Pie colour field',
        )

        multiple_column_field = []
        self._get_query_builder_foreign_key_fields(
            base_model=base_model,
            report_builder_class=report_builder_class,
            fields=multiple_column_field,
        )

        self.fields['multiple_column_field'] = ChoiceField(choices=multiple_column_field, required=False)

        if data_attr.get('multiple_columns') == '1':
            self.fields['multiple_columns'].initial = True
            self.fields['multiple_column_field'].initial = data_attr.get('multiple_column_field')
            if data_attr.get('append_column_title') == '1':
                self.fields['append_column_title'].initial = True

        super().setup_modal(*args, **kwargs)

    def get_additional_attributes(self):
        attributes = []
        self.get_report_type_details()

        attributes.append(f'pie_colour-{self.cleaned_data["pie_colour"]}')

        if self.cleaned_data['has_filter']:
            attributes.append('has_filter-1')

            if self.cleaned_data['filter']:
                b64_filter = encode_attribute(self.cleaned_data['filter'])
                attributes.append(f'filter-{b64_filter}')
            if self.cleaned_data['multiple_columns']:
                attributes.append('multiple_columns-1')
                if self.cleaned_data['positive_colour_field']:
                    attributes.append(f'positive_colour_field-{self.cleaned_data["positive_colour_field"]}')
                attributes.append(f'multiple_column_field-{self.cleaned_data["multiple_column_field"]}')
                if self.cleaned_data['append_column_title']:
                    attributes.append('append_column_title-1')

        if attributes:
            return '-'.join(attributes)
        return None


class PieChartFieldModal(QueryBuilderModalBaseMixin, FormModal):
    form_class = PieChartFieldForm
    size = 'xl'
    template_name = 'advanced_report_builder/charts/modal_field.html'
    no_header_x = True
    helper_class = HorizontalNoEnterHelper

    @property
    def modal_title(self):
        data = json.loads(base64.b64decode(self.slug['data']))
        return f'Edit {data["title"]}'

    def form_valid(self, form):
        selector = self.slug['selector']

        _attr = form.get_additional_attributes()
        self.add_command(
            {
                'function': 'set_attr',
                'selector': f'#{selector}',
                'attr': 'data-attr',
                'val': _attr,
            }
        )

        self.add_command(
            {
                'function': 'html',
                'selector': f'#{selector} span',
                'html': form.cleaned_data['title'],
            }
        )
        self.add_command({'function': 'update_selection'})
        return self.command_response('close')

    # noinspection PyMethodMayBeStatic
    def form_setup(self, form, *_args, **_kwargs):
        form.add_trigger(
            'has_filter',
            'onchange',
            [
                {
                    'selector': '#filter_fields_div',
                    'values': {'checked': 'show'},
                    'default': 'hide',
                }
            ],
        )

        form.add_trigger(
            'multiple_columns',
            'onchange',
            [
                {
                    'selector': '#multiple_columns_fields_div',
                    'values': {'checked': 'show'},
                    'default': 'hide',
                },
            ],
        )

        return [
            'title',
            'pie_colour',
            Div(
                FieldEx(
                    'has_filter',
                    template='django_modals/fields/label_checkbox.html',
                    field_class='col-6 input-group-sm',
                ),
                Div(
                    FieldEx(
                        'filter',
                        template='advanced_report_builder/datatables/fields/single_query_builder.html',
                    ),
                    FieldEx(
                        'multiple_columns',
                        template='django_modals/fields/label_checkbox.html',
                        field_class='col-6 input-group-sm',
                    ),
                    Div(
                        FieldEx(
                            'append_column_title',
                            template='django_modals/fields/label_checkbox.html',
                            field_class='col-6 input-group-sm',
                        ),
                        FieldEx('multiple_column_field'),
                        FieldEx('positive_colour_field'),
                        css_id='multiple_columns_fields_div',
                    ),
                    css_id='filter_fields_div',
                ),
                css_id='annotations_fields_div',
            ),
        ]

    def ajax_get_query_builder_fields(self, **kwargs):
        field_auto_id = kwargs['field_auto_id']

        report_type_id = self.slug['report_type_id']
        query_builder_filters = self.get_query_builder_report_type_field(report_type_id=report_type_id)

        return self.command_response(f'query_builder_{field_auto_id}', data=json.dumps(query_builder_filters))
